"""
Test objects for the sellar two discipline problem.

From Sellar's analytic problem.

    Sellar, R. S., Batill, S. M., and Renaud, J. E., "Response Surface Based, Concurrent Subspace
    Optimization for Multidisciplinary System Design," Proceedings References 79 of the 34th AIAA
    Aerospace Sciences Meeting and Exhibit, Reno, NV, January 1996.
"""
import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.sellar import SellarDis1withDerivatives, \
                         SellarDis2withDerivatives


class SellarDerivatives(om.Group):
    """
    Group containing the Sellar MDA. This version uses the disciplines with derivatives.
    """

    def setup(self):
        self.add_subsystem('d1', SellarDis1withDerivatives(), promotes=['x', 'z', 'y1', 'y2'])
        self.add_subsystem('d2', SellarDis2withDerivatives(), promotes=['z', 'y1', 'y2'])

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)', obj=0.0,
                                                  x=0.0, z=np.array([0.0, 0.0]), y1=0.0, y2=0.0),
                           promotes=['obj', 'x', 'z', 'y1', 'y2'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1', con1=0.0, y1=0.0),
                           promotes=['con1', 'y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0', con2=0.0, y2=0.0),
                           promotes=['con2', 'y2'])

        self.set_input_defaults('x', 1.0)
        self.set_input_defaults('z', np.array([5.0, 2.0]))


class SellarDis1(om.ExplicitComponent):
    """
    Component containing Discipline 1 -- no derivatives version.
    """

    def setup(self):

        # Global Design Variable
        self.add_input('z', val=np.zeros(2))

        # Local Design Variable
        self.add_input('x', val=0.)

        # Coupling parameter
        self.add_input('y2', val=1.0)

        # Coupling output
        self.add_output('y1', val=1.0)

    def setup_partials(self):
        # Finite difference all partials.
        self.declare_partials('*', '*', method='fd')

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y1 = z1**2 + z2 + x1 - 0.2*y2
        """
        z1 = inputs['z'][0]
        z2 = inputs['z'][1]
        x1 = inputs['x']
        y2 = inputs['y2']

        outputs['y1'] = z1**2 + z2 + x1 - 0.2*y2


class SellarDis2(om.ExplicitComponent):
    """
    Component containing Discipline 2 -- no derivatives version.
    """

    def setup(self):
        # Global Design Variable
        self.add_input('z', val=np.zeros(2))

        # Coupling parameter
        self.add_input('y1', val=1.0)

        # Coupling output
        self.add_output('y2', val=1.0)

    def setup_partials(self):
        # Finite difference all partials.
        self.declare_partials('*', '*', method='fd')

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y2 = y1**(.5) + z1 + z2
        """

        z1 = inputs['z'][0]
        z2 = inputs['z'][1]
        y1 = inputs['y1']

        # Note: this may cause some issues. However, y1 is constrained to be
        # above 3.16, so lets just let it converge, and the optimizer will
        # throw it out
        if y1.real < 0.0:
            y1 *= -1

        outputs['y2'] = y1**.5 + z1 + z2


class SellarMDA(om.Group):
    """
    Group containing the Sellar MDA.
    """

    def setup(self):
        cycle = self.add_subsystem('cycle', om.Group(), promotes=['*'])
        cycle.add_subsystem('d1', SellarDis1(), promotes_inputs=['x', 'z', 'y2'],
                            promotes_outputs=['y1'])
        cycle.add_subsystem('d2', SellarDis2(), promotes_inputs=['z', 'y1'],
                            promotes_outputs=['y2'])

        cycle.set_input_defaults('x', 1.0)
        cycle.set_input_defaults('z', np.array([5.0, 2.0]))

        # Nonlinear Block Gauss Seidel is a gradient free solver
        cycle.nonlinear_solver = om.NonlinearBlockGS()

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                  z=np.array([0.0, 0.0]), x=0.0),
                           promotes=['x', 'z', 'y1', 'y2', 'obj'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes=['con1', 'y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes=['con2', 'y2'])


class SellarMDAWithUnits(om.Group):
    """
    Group containing the Sellar MDA.
    """

    class SellarDis1Units(om.ExplicitComponent):
        """
        Component containing Discipline 1 -- no derivatives version.
        """

        def setup(self):

            # Global Design Variable
            self.add_input('z', val=np.zeros(2), units='degC')

            # Local Design Variable
            self.add_input('x', val=0., units='degC')

            # Coupling parameter
            self.add_input('y2', val=1.0, units='degC')

            # Coupling output
            self.add_output('y1', val=1.0, units='degC')

        def setup_partials(self):
            # Finite difference all partials.
            self.declare_partials('*', '*', method='fd')

        def compute(self, inputs, outputs):
            """
            Evaluates the equation
            y1 = z1**2 + z2 + x1 - 0.2*y2
            """
            z1 = inputs['z'][0]
            z2 = inputs['z'][1]
            x1 = inputs['x']
            y2 = inputs['y2']

            outputs['y1'] = z1**2 + z2 + x1 - 0.2*y2


    class SellarDis2Units(om.ExplicitComponent):
        """
        Component containing Discipline 2 -- no derivatives version.
        """

        def setup(self):
            # Global Design Variable
            self.add_input('z', val=np.zeros(2), units='degC')

            # Coupling parameter
            self.add_input('y1', val=1.0, units='degC')

            # Coupling output
            self.add_output('y2', val=1.0, units='degC')

        def setup_partials(self):
            # Finite difference all partials.
            self.declare_partials('*', '*', method='fd')

        def compute(self, inputs, outputs):
            """
            Evaluates the equation
            y2 = y1**(.5) + z1 + z2
            """

            z1 = inputs['z'][0]
            z2 = inputs['z'][1]
            y1 = inputs['y1']

            # Note: this may cause some issues. However, y1 is constrained to be
            # above 3.16, so lets just let it converge, and the optimizer will
            # throw it out
            if y1.real < 0.0:
                y1 *= -1

            outputs['y2'] = y1**.5 + z1 + z2

    def setup(self):

        cycle = self.add_subsystem('cycle', om.Group(), promotes=['*'])
        cycle.add_subsystem('d1', self.SellarDis1Units(), promotes_inputs=['x', 'z', 'y2'],
                            promotes_outputs=['y1'])
        cycle.add_subsystem('d2', self.SellarDis2Units(), promotes_inputs=['z', 'y1'],
                            promotes_outputs=['y2'])

        cycle.set_input_defaults('x', 1.0, units='degC')
        cycle.set_input_defaults('z', np.array([5.0, 2.0]), units='degC')

        # Nonlinear Block Gauss Seidel is a gradient free solver
        cycle.nonlinear_solver = om.NonlinearBlockGS()

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                  z={'val': np.array([0.0, 0.0]), 'units': 'degC'},
                                                  x={'val': 0.0, 'units': 'degC'},
                                                  y1={'units': 'degC'},
                                                  y2={'units': 'degC'}),
                           promotes=['x', 'z', 'y1', 'y2', 'obj'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1', y1={'units': 'degC'},
                                                   con1={'units': 'degC'}),
                           promotes=['con1', 'y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0', y2={'units': 'degC'},
                                                   con2={'units': 'degC'}),
                           promotes=['con2', 'y2'])


class SellarMDALinearSolver(om.Group):
    """
    Group containing the Sellar MDA.
    """

    def setup(self):

        cycle = self.add_subsystem('cycle', om.Group(), promotes=['*'])
        d1 = cycle.add_subsystem('d1', SellarDis1(), promotes_inputs=['x', 'z', 'y2'],
                                 promotes_outputs=['y1'])
        d2 = cycle.add_subsystem('d2', SellarDis2(), promotes_inputs=['z', 'y1'],
                                 promotes_outputs=['y2'])

        self.set_input_defaults('x', 1.0)
        self.set_input_defaults('z', np.array([5.0, 2.0]))

        cycle.nonlinear_solver = om.NonlinearBlockGS()
        cycle.linear_solver = om.DirectSolver()

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                           z=np.array([0.0, 0.0]), x=0.0),
                           promotes=['x', 'z', 'y1', 'y2', 'obj'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes=['con1', 'y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes=['con2', 'y2'])


class SellarDis1CS(om.ExplicitComponent):
    """
    Component containing Discipline 1 -- no derivatives version.
    Uses Complex Step
    """

    def setup(self):

        # Global Design Variable
        self.add_input('z', val=np.zeros(2))

        # Local Design Variable
        self.add_input('x', val=0.)

        # Coupling parameter
        self.add_input('y2', val=1.0)

        # Coupling output
        self.add_output('y1', val=1.0)

    def setup_partials(self):
        # Finite difference all partials.
        self.declare_partials('*', '*', method='cs')

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y1 = z1**2 + z2 + x1 - 0.2*y2
        """
        z1 = inputs['z'][0]
        z2 = inputs['z'][1]
        x1 = inputs['x']
        y2 = inputs['y2']

        outputs['y1'] = z1**2 + z2 + x1 - 0.2*y2


class SellarDis2CS(om.ExplicitComponent):
    """
    Component containing Discipline 2 -- no derivatives version.
    Uses Complex Step
    """

    def setup(self):
        # Global Design Variable
        self.add_input('z', val=np.zeros(2))

        # Coupling parameter
        self.add_input('y1', val=1.0)

        # Coupling output
        self.add_output('y2', val=1.0)

    def setup_partials(self):
        # Finite difference all partials.
        self.declare_partials('*', '*', method='cs')

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y2 = y1**(.5) + z1 + z2
        """

        z1 = inputs['z'][0]
        z2 = inputs['z'][1]
        y1 = inputs['y1']

        # Note: this may cause some issues. However, y1 is constrained to be
        # above 3.16, so lets just let it converge, and the optimizer will
        # throw it out
        if y1.real < 0.0:
            y1 *= -1

        outputs['y2'] = y1**.5 + z1 + z2


class SellarNoDerivativesCS(om.Group):
    """
    Group containing the Sellar MDA. This version uses the disciplines without derivatives.
    """

    def setup(self):
        self.add_subsystem('px', om.IndepVarComp('x', 1.0), promotes=['x'])
        self.add_subsystem('pz', om.IndepVarComp('z', np.array([5.0, 2.0])), promotes=['z'])

        cycle = self.add_subsystem('cycle', om.Group(), promotes=['x', 'z', 'y1', 'y2'])
        d1 = cycle.add_subsystem('d1', SellarDis1CS(), promotes=['x', 'z', 'y1', 'y2'])
        d2 = cycle.add_subsystem('d2', SellarDis2CS(), promotes=['z', 'y1', 'y2'])

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                                                  z=np.array([0.0, 0.0]), x=0.0),
                           promotes=['x', 'z', 'y1', 'y2', 'obj'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes=['con1', 'y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes=['con2', 'y2'])

        self.nonlinear_solver = om.NonlinearBlockGS()
        self.linear_solver = om.ScipyKrylov()


class SellarIDF(om.Group):
    """
    Individual Design Feasible (IDF) architecture for the Sellar problem.
    """
    def setup(self):
        # construct the Sellar model with `y1` and `y2` as independent variables

        self.set_input_defaults('x', 5.)
        self.set_input_defaults('y1', 5.)
        self.set_input_defaults('y2', 5.)
        self.set_input_defaults('z', np.array([2., 0.]))

        self.add_subsystem('d1', SellarDis1withDerivatives(), promotes_inputs=['x', 'z', 'y2'])
        self.add_subsystem('d2', SellarDis2withDerivatives(), promotes_inputs=['y1', 'z'])

        self.add_subsystem('obj_cmp', om.ExecComp('obj = x**2 + z[1] + y1 + exp(-y2)',
                           x=0., z=np.array([0., 0.])), promotes_inputs=['x', 'z', 'y1', 'y2'])

        self.add_subsystem('con_cmp1', om.ExecComp('con1 = 3.16 - y1'), promotes_inputs=['y1'])
        self.add_subsystem('con_cmp2', om.ExecComp('con2 = y2 - 24.0'), promotes_inputs=['y2'])

        # rather than create a cycle by connecting d1.y1 to d2.y1 and d2.y2 to d1.y2
        # we will constrain y1 and y2 to be equal for the two disciplines

        equal = om.EQConstraintComp()
        self.add_subsystem('equal', equal, promotes_inputs=[('lhs:y1', 'y1'), ('lhs:y2', 'y2')])

        equal.add_eq_output('y1', add_constraint=True)
        equal.add_eq_output('y2', add_constraint=True)

        self.connect('d1.y1', 'equal.rhs:y1')
        self.connect('d2.y2', 'equal.rhs:y2')

        # the driver will effectively solve the cycle
        # by satisfying the equality constraints

        self.add_design_var('x', lower=0., upper=5.)
        self.add_design_var('y1', lower=0., upper=5.)
        self.add_design_var('y2', lower=0., upper=5.)
        self.add_design_var('z', lower=np.array([-5., 0.]), upper=np.array([5., 5.]))
        self.add_objective('obj_cmp.obj')
        self.add_constraint('con_cmp1.con1', upper=0.)
        self.add_constraint('con_cmp2.con2', upper=0.)



import numpy as np
from math import pi
import openmdao.api as om

# Model
interp = om.MetaModelUnStructuredComp()

# Training Data
x_train1 = np.random.uniform(0, pi, 100)
x_train2 = np.random.uniform(0, pi, 100)
x_train3 = np.random.uniform(0, pi, 100)
y = np.sin(x_train1 * x_train2 * x_train3)

# Inputs
interp.add_input('input_1', 0., training_data=x_train1)
interp.add_input('input_2', 0., training_data=x_train2)
interp.add_input('input_3', 0., training_data=x_train3)

# Outputs
interp.add_output('output_1', 0., training_data=y)

# Surrogate Model
interp.options['default_surrogate'] = om.KrigingSurrogate()

prob = om.Problem()
prob.model.add_subsystem('interp', interp)
prob.setup()
prob.final_setup()


""" Unit tests for semi structured metamodels in view_mm. """
import unittest

import numpy as np

try:
    import bokeh
    from openmdao.visualization.meta_model_viewer.meta_model_visualization import MetaModelVisualization
except ImportError:
    bokeh = None

import openmdao.api as om


@unittest.skipUnless(bokeh, "Bokeh is required")
class SemistructuredMetaModelCompTests(unittest.TestCase):

    def test_basic(self):
        # Tests that semi structured grids load without error.

        grid = np.array([
            [1.0, 5.0, 8.0],
            [1.0, 5.0, 9.0],
            [1.0, 5.0, 10.0],
            [1.0, 5.0, 20.0],
            [1.0, 5.3, 8.0],
            [1.0, 5.3, 9.0],
            [1.0, 5.3, 10.0],
            [1.0, 5.3, 20.0],
            [1.0, 5.6, 8.0],
            [1.0, 5.6, 9.0],
            [1.0, 5.6, 10.0],
            [1.0, 5.6, 20.0],
            [1.0, 6.0, 8.0],
            [1.0, 6.0, 9.0],
            [1.0, 6.0, 10.0],
            [1.0, 6.0, 20.0],
            [2.0, 7.0, 13.0],
            [2.0, 7.0, 14.0],
            [2.0, 7.0, 15.0],
            [2.0, 7.0, 16.0],
            [2.0, 8.0, 13.0],
            [2.0, 8.0, 14.0],
            [2.0, 8.0, 15.0],
            [2.0, 8.0, 16.0],
            [2.0, 8.5, 13.0],
            [2.0, 8.5, 14.0],
            [2.0, 8.5, 15.0],
            [2.0, 8.5, 16.0],
            [2.0, 9.0, 13.0],
            [2.0, 9.0, 14.0],
            [2.0, 9.0, 15.0],
            [2.0, 9.0, 16.0],
        ])

        values = 15.0 + 2 * np.random.random(32)

        prob = om.Problem()
        model = prob.model

        interp = om.MetaModelSemiStructuredComp(vec_size=3, training_data_gradients=True)
        interp.add_input('x', training_data=grid[:, 0])
        interp.add_input('y', training_data=grid[:, 1])
        interp.add_input('z', training_data=grid[:, 2])
        interp.add_output('f', training_data=values)

        model.add_subsystem('interp', interp)

        prob.setup(force_alloc_complex=True)
        prob.run_model()

        viz = MetaModelVisualization(interp)


if __name__ == '__main__':
    unittest.main()



"""Some basic shell utilities, used for ExternalCodeComp mostly."""
import os
import signal
import subprocess
import sys
import time

PIPE = subprocess.PIPE
STDOUT = subprocess.STDOUT
DEV_NULL = 'nul:' if sys.platform == 'win32' else '/dev/null'


class CalledProcessError(subprocess.CalledProcessError):
    """
    :class:`subprocess.CalledProcessError` plus `errormsg` attribute.

    Parameters
    ----------
    returncode : int
        Error code for this error.
    cmd : str or list
        If a string, then this is the command line to execute, and the
        :class:`subprocess.Popen` ``shell`` argument is set True.
        Otherwise, this is a list of arguments; the first is the command
        to execute.
    errormsg : str
        Error message for this error.

    Attributes
    ----------
    errormsg : str
        Error message saved for string access.
    """

    def __init__(self, returncode, cmd, errormsg):
        """
        Initialize.
        """
        super().__init__(returncode, cmd)
        self.errormsg = errormsg

    def __str__(self):
        """
        Return string of error message.

        Returns
        -------
        str
            Error message.
        """
        return 'Command %r returned non-zero exit status %d: %s' % (self.cmd, self.returncode,
                                                                    self.errormsg)


class ShellProc(subprocess.Popen):
    """
    A slight modification to :class:`subprocess.Popen`.

    If `args` is a string, then the ``shell`` argument is set True.
    Updates a copy of ``os.environ`` with `env` and opens files for any
    stream which is a :class:`str`.

    Parameters
    ----------
    args : str or list
        If a string, then this is the command line to execute and the
        :class:`subprocess.Popen` ``shell`` argument is set True.
        Otherwise, this is a list of arguments; the first is the command
        to execute.
    stdin : str, file, or int
        Specify handling of stdin stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stdout : str, file, or int
        Specify handling of stdout stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stderr : str, file, or int
        Specify handling of stderr stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    env : dict
        Environment variables for the command.
    universal_newlines : bool
        Set to True to turn on universal newlines.

    Attributes
    ----------
    _stdin_arg : str, file, or int
        Save handle to make closing easier.
    _stdout_arg : str, file, or int
        Save handle to make closing easier.
    _stderr_arg : str, file, or int
        Save handle to make closing easier.
    _inp : str, file, or int
        Save handle to make closing easier.
    _out : str, file, or int
        Save handle to make closing easier.
    _err : str, file, or int
        Save handle to make closing easier.
    """

    def __init__(self, args, stdin=None, stdout=None, stderr=None, env=None,
                 universal_newlines=False):
        """
        Initialize.
        """
        environ = os.environ.copy()
        if env:
            environ.update(env)

        self._stdin_arg = stdin
        self._stdout_arg = stdout
        self._stderr_arg = stderr

        if isinstance(stdin, str):
            self._inp = open(stdin, 'r')
        else:
            self._inp = stdin

        if isinstance(stdout, str):
            self._out = open(stdout, 'w')
        else:
            self._out = stdout

        if isinstance(stderr, str):
            self._err = open(stderr, 'w')
        else:
            self._err = stderr

        shell = isinstance(args, str)

        try:
            if sys.platform == 'win32':
                subprocess.Popen.__init__(self, args, stdin=self._inp,
                                          stdout=self._out, stderr=self._err,
                                          shell=shell, env=environ,  # nosec: user responsibility
                                          universal_newlines=universal_newlines)
            else:
                subprocess.Popen.__init__(self, args, stdin=self._inp,
                                          stdout=self._out, stderr=self._err,
                                          shell=shell, env=environ,  # nosec: user responsibility
                                          universal_newlines=universal_newlines,
                                          # setsid to put this and any children in
                                          # same process group so we can kill them
                                          # all if necessary
                                          preexec_fn=os.setsid)

        except Exception:
            self.close_files()
            raise

    def close_files(self):
        """
        Close files that were implicitly opened.
        """
        if isinstance(self._stdin_arg, str):
            self._inp.close()
        if isinstance(self._stdout_arg, str):
            self._out.close()
        if isinstance(self._stderr_arg, str):
            self._err.close()

    def terminate(self, timeout=None):
        """
        Stop child process.

        If `timeout` is specified, then :meth:`wait` will be called to wait for the process
        to terminate.

        Parameters
        ----------
        timeout : float (seconds)
            Maximum time to wait for the process to stop.
            A value of zero implies an infinite maximum wait.

        Returns
        -------
        int
            Return Code.
        str
            Error Message.
        """
        if sys.platform == 'win32':
            subprocess.Popen(f"TASKKILL /F /PID {self.pid} /T")  # nosec: trusted input
        else:
            os.killpg(os.getpgid(self.pid), signal.SIGTERM)

        if timeout is not None:
            return self.wait(timeout=timeout)

    def wait(self, poll_delay=0., timeout=0.):
        """
        Poll for command completion or timeout.

        Closes any files implicitly opened.

        Parameters
        ----------
        poll_delay : float (seconds)
            Time to delay between polling for command completion.
            A value of zero uses an internal default.
        timeout : float (seconds)
            Maximum time to wait for command completion.
            A value of zero implies an infinite maximum wait.

        Returns
        -------
        int
            Return Code.
        str
            Error Message.
        """
        return_code = None
        try:
            if poll_delay <= 0:
                poll_delay = max(0.1, timeout / 100.)
                poll_delay = min(10., poll_delay)
            npolls = int(timeout / poll_delay) + 1

            time.sleep(poll_delay)
            return_code = self.poll()
            while return_code is None:
                npolls -= 1
                if (timeout > 0) and (npolls < 0):
                    self.terminate()
                    break
                time.sleep(poll_delay)
                return_code = self.poll()
        finally:
            self.close_files()

        # self.returncode set by self.poll().
        if return_code is not None:
            self.errormsg = self.error_message(return_code)
        else:
            self.errormsg = 'Timed out'
        return (return_code, self.errormsg)

    def error_message(self, return_code):
        """
        Return error message for `return_code`.

        The error messages are derived from the operating system definitions.
        Some programs don't necessarily return exit codes conforming to these
        definitions.

        Parameters
        ----------
        return_code : int
            Return code from :meth:`poll`.

        Returns
        -------
        str
            Error Message string.
        """
        error_msg = ''
        if return_code:
            if return_code > 0:
                try:
                    err_msg = os.strerror(return_code)
                except OverflowError:
                    err_msg = "Process exited with unknown return code {}".format(return_code)
            elif sys.platform != 'win32':
                sig = -return_code
                if sig < signal.NSIG:
                    for item in signal.__dict__.keys():
                        if item.startswith('SIG'):
                            if getattr(signal, item) == sig:
                                error_msg = ': %s' % item
                                break

        return error_msg


def call(args, stdin=None, stdout=None, stderr=None, env=None,
         poll_delay=0., timeout=0.):
    """
    Run command with arguments.

    Parameters
    ----------
    args : str or list
        If a string, then this is the command line to execute and the
        :class:`subprocess.Popen` ``shell`` argument is set True.
        Otherwise, this is a list of arguments; the first is the command
        to execute.
    stdin : str, file, or int
        Specify handling of stdin stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stdout : str, file, or int
        Specify handling of stdout stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stderr : str, file, or int
        Specify handling of stderr stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    env : dict
        Environment variables for the command.
    poll_delay : float (seconds)
        Time to delay between polling for command completion.
        A value of zero uses an internal default.
    timeout : float (seconds)
        Maximum time to wait for command completion.
        A value of zero implies an infinite maximum wait.

    Returns
    -------
    int
        Return Code.
    str
        Error Message.
    """
    process = ShellProc(args, stdin, stdout, stderr, env)
    return process.wait(poll_delay, timeout)


def check_call(args, stdin=None, stdout=None, stderr=None, env=None,
               poll_delay=0., timeout=0.):
    """
    Run command with arguments.

    Raises :class:`CalledProcessError` if process returns an error code.

    Parameters
    ----------
    args : str or list
        If a string, then this is the command line to execute, and the
        :class:`subprocess.Popen` ``shell`` argument is set True.
        Otherwise, this is a list of arguments; the first is the command
        to execute.
    stdin : str, file, or int
        Specify handling of stdin stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stdout : str, file, or int
        Specify handling of stdout stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    stderr : str, file, or int
        Specify handling of stderr stream. If a string, a file
        of that name is opened. Otherwise, see the :mod:`subprocess`
        documentation.
    env : dict
        Environment variables for the command.
    poll_delay : float (seconds)
        Time to delay between polling for command completion.
        A value of zero uses an internal default.
    timeout : float (seconds)
        Maximum time to wait for command completion.
        A value of zero implies an infinite maximum wait.
    """
    process = ShellProc(args, stdin, stdout, stderr, env)
    return_code, error_msg = process.wait(poll_delay, timeout)
    if return_code:
        raise CalledProcessError(return_code, args, error_msg)



"""
Unit test for DotProductComp.
"""
import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import force_check_partials


class TestDotProductCompNx3(unittest.TestCase):

    def setUp(self):
        self.nn = 5

        self.p = om.Problem()

        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(self.nn, 3))
        ivc.add_output(name='b', shape=(self.nn, 3))

        self.p.model.add_subsystem(name='ivc',
                                   subsys=ivc,
                                   promotes_outputs=['a', 'b'])

        self.p.model.add_subsystem(name='dot_prod_comp',
                                   subsys=om.DotProductComp(vec_size=self.nn))

        self.p.model.connect('a', 'dot_prod_comp.a')
        self.p.model.connect('b', 'dot_prod_comp.b')

        self.p.setup()

        self.p['a'] = np.random.rand(self.nn, 3)
        self.p['b'] = np.random.rand(self.nn, 3)

        self.p.run_model()

    def test_results(self):

        for i in range(self.nn):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p['dot_prod_comp.c'][i]
            expected_i = np.dot(a_i, b_i)

            np.testing.assert_almost_equal(c_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestDotProductCompNx4(unittest.TestCase):
    def setUp(self):
        self.nn = 100

        self.p = om.Problem()

        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(self.nn, 4))
        ivc.add_output(name='b', shape=(self.nn, 4))

        self.p.model.add_subsystem(name='ivc',
                                   subsys=ivc,
                                   promotes_outputs=['a', 'b'])

        self.p.model.add_subsystem(name='dot_prod_comp',
                                   subsys=om.DotProductComp(vec_size=self.nn, length=4))

        self.p.model.connect('a', 'dot_prod_comp.a')
        self.p.model.connect('b', 'dot_prod_comp.b')

        self.p.setup()

        self.p['a'] = np.random.rand(self.nn, 4)
        self.p['b'] = np.random.rand(self.nn, 4)

        self.p.run_model()

    def test_results(self):

        for i in range(self.nn):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p['dot_prod_comp.c'][i]
            expected_i = np.dot(a_i, b_i)
            np.testing.assert_almost_equal(c_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestUnits(unittest.TestCase):

    def setUp(self):
        self.nn = 5

        self.p = om.Problem()

        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(self.nn, 3), units='lbf')
        ivc.add_output(name='b', shape=(self.nn, 3), units='ft/s')

        self.p.model.add_subsystem(name='ivc',
                                   subsys=ivc,
                                   promotes_outputs=['a', 'b'])

        self.p.model.add_subsystem(name='dot_prod_comp',
                                   subsys=om.DotProductComp(vec_size=self.nn,
                                                            a_units='N', b_units='m/s',
                                                            c_units='W'))

        self.p.model.connect('a', 'dot_prod_comp.a')
        self.p.model.connect('b', 'dot_prod_comp.b')

        self.p.setup()

        self.p['a'] = np.random.rand(self.nn, 3)
        self.p['b'] = np.random.rand(self.nn, 3)

        self.p.run_model()

    def test_results(self):

        for i in range(self.nn):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p.get_val('dot_prod_comp.c', units='hp')[i]
            expected_i = np.dot(a_i, b_i) / 550.0

            np.testing.assert_almost_equal(c_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestMultipleUnits(unittest.TestCase):

    def setUp(self):
        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(5, 3), units='lbf')
        ivc.add_output(name='b', shape=(5, 3), units='ft/s')
        ivc.add_output(name='x', shape=(3, 7), units='N')
        ivc.add_output(name='y', shape=(3, 7), units='m/s')

        dpc = om.DotProductComp(vec_size=5,  # default length=3
                                a_units='N', b_units='m/s', c_units='W')

        dpc.add_product(a_name='x', b_name='y', c_name='z', vec_size=3, length=7,
                        a_units='N', b_units='m/s', c_units='hp')

        model = om.Group()
        model.add_subsystem(name='ivc', subsys=ivc,
                            promotes_outputs=['a', 'b', 'x', 'y'])

        model.add_subsystem(name='dot_prod_comp', subsys=dpc)

        model.connect('a', 'dot_prod_comp.a')
        model.connect('b', 'dot_prod_comp.b')
        model.connect('x', 'dot_prod_comp.x')
        model.connect('y', 'dot_prod_comp.y')

        p = self.p = om.Problem(model)
        p.setup()

        p['a'] = np.random.rand(5, 3)
        p['b'] = np.random.rand(5, 3)
        p['x'] = np.random.rand(3, 7)
        p['y'] = np.random.rand(3, 7)

        p.run_model()

    def test_results(self):

        for i in range(5):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p.get_val('dot_prod_comp.c', units='hp')[i]
            expected_i = np.dot(a_i, b_i) / 550.0

            np.testing.assert_almost_equal(c_i, expected_i)

        for i in range(3):
            x_i = self.p['x'][i, :]
            y_i = self.p['y'][i, :]
            z_i = self.p.get_val('dot_prod_comp.z', units='hp')[i]
            expected_i = np.dot(x_i, y_i)
            np.testing.assert_almost_equal(z_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestMultipleConfigure(unittest.TestCase):

    def setUp(self):

        class MyModel(om.Group):

            def setup(self):
                ivc = om.IndepVarComp()
                ivc.add_output(name='a', shape=(5, 3), units='lbf')
                ivc.add_output(name='b', shape=(5, 3), units='ft/s')

                dpc = om.DotProductComp(vec_size=5,  # default length=3
                                        a_units='N', b_units='m/s', c_units='W')

                self.add_subsystem('ivc', ivc, promotes_outputs=['*'])
                self.add_subsystem('dpc', dpc)

                self.connect('a', 'dpc.a')
                self.connect('b', 'dpc.b')

            def configure(self):
                self.ivc.add_output(name='x', shape=(3, 7), units='N')
                self.ivc.add_output(name='y', shape=(3, 7), units='m/s')

                self.dpc.add_product(a_name='x', b_name='y', c_name='z', vec_size=3, length=7,
                                     a_units='N', b_units='m/s', c_units='hp')


                self.connect('x', 'dpc.x')
                self.connect('y', 'dpc.y')

        p = self.p = om.Problem(MyModel())
        p.setup()

        p['a'] = np.random.rand(5, 3)
        p['b'] = np.random.rand(5, 3)
        p['x'] = np.random.rand(3, 7)
        p['y'] = np.random.rand(3, 7)

        p.run_model()

    def test_results(self):

        for i in range(5):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p.get_val('dpc.c', units='hp')[i]
            expected_i = np.dot(a_i, b_i) / 550.0

            np.testing.assert_almost_equal(c_i, expected_i)

        for i in range(3):
            x_i = self.p['x'][i, :]
            y_i = self.p['y'][i, :]
            z_i = self.p.get_val('dpc.z', units='hp')[i]
            expected_i = np.dot(x_i, y_i)
            np.testing.assert_almost_equal(z_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestMultipleCommonA(unittest.TestCase):

    def setUp(self):
        self.nn = 5

        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(self.nn, 3), units='lbf')
        ivc.add_output(name='b', shape=(self.nn, 3), units='ft/s')
        ivc.add_output(name='y', shape=(self.nn, 3), units='ft/s')

        dpc = om.DotProductComp(vec_size=self.nn,  # default length=3
                                a_units='N', b_units='m/s', c_units='W')

        dpc.add_product('z', b_name='y', vec_size=self.nn, length=3,
                        a_units='N', b_units='m/s', c_units='W')

        model = om.Group()
        model.add_subsystem(name='ivc', subsys=ivc,
                            promotes_outputs=['a', 'b', 'y'])

        model.add_subsystem(name='dot_prod_comp', subsys=dpc)

        model.connect('a', 'dot_prod_comp.a')
        model.connect('b', 'dot_prod_comp.b')
        model.connect('y', 'dot_prod_comp.y')

        p = self.p = om.Problem(model)
        p.setup()

        p['a'] = np.random.rand(self.nn, 3)
        p['b'] = np.random.rand(self.nn, 3)
        p['y'] = np.random.rand(self.nn, 3)

        p.run_model()

    def test_results(self):

        for i in range(self.nn):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p.get_val('dot_prod_comp.c', units='hp')[i]
            expected_i = np.dot(a_i, b_i) / 550.

            np.testing.assert_almost_equal(c_i, expected_i)

            y_i = self.p['y'][i, :]
            z_i = self.p.get_val('dot_prod_comp.z', units='hp')[i]
            expected_i = np.dot(a_i, y_i) / 550.
            np.testing.assert_almost_equal(z_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestMultipleCommonB(unittest.TestCase):

    def setUp(self):
        self.nn = 5

        ivc = om.IndepVarComp()
        ivc.add_output(name='a', shape=(self.nn, 3), units='lbf')
        ivc.add_output(name='b', shape=(self.nn, 3), units='ft/s')
        ivc.add_output(name='x', shape=(self.nn, 3), units='lbf')

        dpc = om.DotProductComp(vec_size=self.nn,  # default length=3
                                a_units='N', b_units='m/s', c_units='W')

        dpc.add_product('z', a_name='x', vec_size=self.nn, length=3,
                        a_units='N', b_units='m/s', c_units='W')

        model = om.Group()
        model.add_subsystem(name='ivc', subsys=ivc,
                            promotes_outputs=['a', 'b', 'x'])

        model.add_subsystem(name='dot_prod_comp', subsys=dpc)

        model.connect('a', 'dot_prod_comp.a')
        model.connect('b', 'dot_prod_comp.b')
        model.connect('x', 'dot_prod_comp.x')

        p = self.p = om.Problem(model)
        p.setup()

        p['a'] = np.random.rand(self.nn, 3)
        p['b'] = np.random.rand(self.nn, 3)
        p['x'] = np.random.rand(self.nn, 3)

        p.run_model()

    def test_results(self):

        for i in range(self.nn):
            a_i = self.p['a'][i, :]
            b_i = self.p['b'][i, :]
            c_i = self.p.get_val('dot_prod_comp.c', units='hp')[i]
            expected_i = np.dot(a_i, b_i) / 550.

            np.testing.assert_almost_equal(c_i, expected_i)

            x_i = self.p['x'][i, :]
            z_i = self.p.get_val('dot_prod_comp.z', units='hp')[i]
            expected_i = np.dot(x_i, b_i) / 550.
            np.testing.assert_almost_equal(z_i, expected_i)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        cpd = force_check_partials(self.p, compact_print=True, out_stream=None)

        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=6)


class TestMultipleErrors(unittest.TestCase):

    def test_duplicate_outputs(self):
        dpc = om.DotProductComp()

        with self.assertRaises(NameError) as ctx:
            dpc.add_product('c')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Multiple definition of output 'c'.")

    def test_input_as_output(self):
        dpc = om.DotProductComp()

        with self.assertRaises(NameError) as ctx:
            dpc.add_product('a', 'b', 'c')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: 'a' specified as "
                         "an output, but it has already been defined as an input.")

    def test_output_as_input_a(self):
        dpc = om.DotProductComp()

        with self.assertRaises(NameError) as ctx:
            dpc.add_product('z', 'c', 'b')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: 'c' specified as "
                         "an input, but it has already been defined as an output.")

    def test_output_as_input_b(self):
        dpc = om.DotProductComp()

        with self.assertRaises(NameError) as ctx:
            dpc.add_product('z', 'b', 'c')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: 'c' specified as "
                         "an input, but it has already been defined as an output.")

    def test_a_vec_size_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'a', 'y', vec_size=10)

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting vec_size=10 specified for input 'a', "
                         "which has already been defined with vec_size=1.")

    def test_a_length_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'a', 'y', length=10)

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting length=10 specified for input 'a', "
                         "which has already been defined with length=3.")

    def test_a_units_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'a', 'b',a_units='ft')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting units 'ft' specified for input 'a', "
                         "which has already been defined with units 'None'.")

    def test_b_vec_size_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'x', 'b', vec_size=10)

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting vec_size=10 specified for input 'b', "
                         "which has already been defined with vec_size=1.")

    def test_b_length_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'x', 'b', length=10)

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting length=10 specified for input 'b', "
                         "which has already been defined with length=3.")

    def test_b_units_mismatch(self):
        dpc = om.DotProductComp()

        with self.assertRaises(ValueError) as ctx:
            dpc.add_product('z', 'a', 'b', b_units='ft')

        self.assertEqual(str(ctx.exception), "<class DotProductComp>: "
                         "Conflicting units 'ft' specified for input 'b', "
                         "which has already been defined with units 'None'.")


class TestFeature(unittest.TestCase):

    def test(self):
        """
        A simple example to compute power as the dot product of force and velocity vectors
        at 24 points simultaneously.
        """

        n = 24

        p = om.Problem()

        dp_comp = om.DotProductComp(vec_size=n, length=3, a_name='F', b_name='v', c_name='P',
                                    a_units='N', b_units='m/s', c_units='W')

        p.model.add_subsystem(name='dot_prod_comp', subsys=dp_comp,
                             promotes_inputs=[('F', 'force'), ('v', 'vel')])

        p.setup()

        p.set_val('force', np.random.rand(n, 3))
        p.set_val('vel', np.random.rand(n, 3))

        p.run_model()

        # Verify the results against numpy.dot in a for loop.
        expected = []
        for i in range(n):
            a_i = p.get_val('force')[i, :]
            b_i = p.get_val('vel')[i, :]
            expected.append(np.dot(a_i, b_i))

            actual_i = p.get_val('dot_prod_comp.P')[i]
            rel_error = np.abs(expected[i] - actual_i)/actual_i
            assert rel_error < 1e-9, f"Relative error: {rel_error}"

        assert_near_equal(p.get_val('dot_prod_comp.P', units='kW'), np.array(expected)/1000.)

    def test_multiple(self):
        """
        Simultaneously compute work as the dot product of force and displacement vectors
        and power as the dot product of force and velocity vectors at 24 points.
        """

        n = 24

        p = om.Problem()

        dp_comp = om.DotProductComp(vec_size=n, length=3,
                                    a_name='F', b_name='d', c_name='W',
                                    a_units='N', b_units='m', c_units='J')

        dp_comp.add_product(vec_size=n, length=3,
                            a_name='F', b_name='v', c_name='P',
                            a_units='N', b_units='m/s', c_units='W')

        p.model.add_subsystem(name='dot_prod_comp', subsys=dp_comp,
                              promotes_inputs=[('F', 'force'), ('d', 'disp'), ('v', 'vel')])

        p.setup()

        p.set_val('force', np.random.rand(n, 3))
        p.set_val('disp', np.random.rand(n, 3))
        p.set_val('vel', np.random.rand(n, 3))

        p.run_model()

        # Verify the results against numpy.dot in a for loop.
        expected_P = []
        expected_W = []
        for i in range(n):
            a_i = p.get_val('force')[i, :]

            b_i = p.get_val('disp')[i, :]
            expected_W.append(np.dot(a_i, b_i))

            actual_i = p.get_val('dot_prod_comp.W')[i]
            rel_error = np.abs(actual_i - expected_W[i])/actual_i
            assert rel_error < 1e-9, f"Relative error: {rel_error}"

            b_i = p.get_val('vel')[i, :]
            expected_P.append(np.dot(a_i, b_i))

            actual_i = p.get_val('dot_prod_comp.P')[i]
            rel_error = np.abs(expected_P[i] - actual_i)/actual_i
            assert rel_error < 1e-9, f"Relative error: {rel_error}"

        assert_near_equal(p.get_val('dot_prod_comp.W', units='kJ'), np.array(expected_W)/1000.)
        assert_near_equal(p.get_val('dot_prod_comp.P', units='kW'), np.array(expected_P)/1000.)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import math
from html.parser import HTMLParser

import openmdao.api as om
from openmdao.visualization.tables.table_builder import generate_table
from openmdao.utils.testing_utils import use_tempdirs


class HTMLTableParser(HTMLParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.found_table = False
        self.table_finished = False
        self.rowcount = 0
        self.headers = []
        self.rows = []
        self.stack = []
        self.current_row = None

    def check_stack(self, tag):
        top = self.stack[-1][0] if self.stack else None
        if top != tag:
            raise RuntimeError(f"Expected '{tag}' at top of stack but found '{top}'.")
        return self.stack[-1]

    def handle_starttag(self, tag, attrs):
        if tag == 'table':
            self.found_table = True
            self.stack.append([tag, None])
        elif tag in ('td', 'th'):
            self.stack.append([tag, None])
        elif tag == 'tr':
            self.stack.append([tag, None, None])

    def handle_endtag(self, tag):
        if tag in ('td', 'th'):
            top = self.check_stack(tag)
            data = top[1]
            self.stack.pop()
            top = self.check_stack('tr')
            if top[1] is None:
                top[1] = []
                top[2] = tag
            elif top[2] != tag:
                raise RuntimeError("Mixed <tr> and <td> on same row!")
            top[1].append(data)
        elif tag == 'tr':
            top = self.check_stack(tag)
            if top[2] == 'td':
                self.rows.append(top[1])
            elif top[2] == 'th':
                self.headers = top[1]
            else:
                raise RuntimeError("Found empty <tr>")
            self.stack.pop()
        elif tag == 'table':
            self.table_finished = True
            if not self.stack or self.stack[-1][0] != 'table':
                raise RuntimeError("</table> doesn't have matching <table>")
            self.stack.pop()

    def handle_data(self, data):
        if self.stack and self.stack[-1][0] in ('td', 'th'):
            self.stack[-1][1] = data


@use_tempdirs
class TableTestCase(unittest.TestCase):
    def check_html(self, rows, headers, expected_rows, expected_headers, **kwargs):
        table = generate_table(rows, tablefmt='html', headers=headers, **kwargs)
        parser = HTMLTableParser()
        parser.feed(str(table))
        self.assertEqual(parser.headers, expected_headers)
        self.assertEqual(parser.rows, expected_rows)

    def check_text(self, fmt, rows, headers, expected, **kwargs):
        table = generate_table(rows, tablefmt=fmt, headers=headers, **kwargs)
        tstr = str(table)
        self.assertEqual(tstr.strip(), expected.strip())

    def table_row_iter(self, c1typ, c2typ, c3typ):
        def yield_ints(n):
            for i in range(n):
                yield i
        def yield_floats(n):
            for i in range(n):
                yield math.pi ** i
        def yield_strs(n):
            base = ['foobar blah', 'asdfas dffff', 'hello world blah blah']
            mult = n // len(base)
            strs = (base * (mult + 1))[:n]
            for s in strs:
                yield s
        def yield_nones(n):
            for i in range(n):
                yield None

        tdict = {
            'int': yield_ints,
            'float': yield_floats,
            'str': yield_strs,
            None: yield_nones,
        }
        colgens = [tdict[c1typ](3), tdict[c2typ](3), tdict[c3typ](3)]
        for i in range(3):
            row = [next(cg) for cg in colgens]
            yield row


    def test_no_header_text(self):
        expected = """
| --------- |
| 0 | 1 | 2 |
| 3 | 4 | 5 |
| 6 | 7 | 8 |
| --------- |"""
        self.check_text('text', np.arange(9).reshape((3,3)), None, expected)
        self.check_text('text', [[0,1,2],[3,4,5],[6,7,8]], None, expected)
        self.check_text('text', iter(np.arange(9).reshape((3,3))), None, expected)

    def test_no_header_github(self):
        expected = """
| 0 | 1 | 2 |
| 3 | 4 | 5 |
| 6 | 7 | 8 |"""
        self.check_text('github', np.arange(9).reshape((3,3)), None, expected)
        self.check_text('github', [[0,1,2],[3,4,5],[6,7,8]], None, expected)

    def test_no_header_rst(self):
        expected = """
=  =  =
0  1  2
3  4  5
6  7  8
=  =  ="""
        self.check_text('rst', np.arange(9).reshape((3,3)), None, expected)
        self.check_text('rst', [[0,1,2],[3,4,5],[6,7,8]], None, expected)

    def test_w_header_text(self):
        headers = [f"Col{i}" for i in range(3)]
        expected = """
| ------------------ |
| Col0 | Col1 | Col2 |
| ---- | ---- | ---- |
|    0 |    1 |    2 |
|    3 |    4 |    5 |
|    6 |    7 |    8 |
| ------------------ |"""
        self.check_text('text', np.arange(9).reshape((3,3)), headers, expected)
        self.check_text('text', [[0,1,2],[3,4,5],[6,7,8]], headers, expected)

    def test_w_header_rst(self):
        headers = [f"Col{i}" for i in range(3)]
        expected = """
====  ====  ====
Col0  Col1  Col2
====  ====  ====
   0     1     2
   3     4     5
   6     7     8
====  ====  ====
"""
        self.check_text('rst', np.arange(9).reshape((3,3)), headers, expected)
        self.check_text('rst', [[0,1,2],[3,4,5],[6,7,8]], headers, expected)

    def test_w_header_github(self):
        headers = [f"Col{i}" for i in range(3)]
        expected = """
| Col0 | Col1 | Col2 |
| ---: | ---: | ---: |
|    0 |    1 |    2 |
|    3 |    4 |    5 |
|    6 |    7 |    8 |
"""
        self.check_text('github', np.arange(9).reshape((3,3)), headers, expected)
        self.check_text('github', [[0,1,2],[3,4,5],[6,7,8]], headers, expected)

    def test_no_header_html(self):
        self.check_html(np.arange(9).reshape((3,3)), None, [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], [])
        self.check_html([[0,1,2],[3,4,5],[6,7,8]], None, [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], [])

    def test_w_header_html(self):
        headers = [f"Col{i}" for i in range(3)]
        self.check_html(np.arange(9).reshape((3,3)), headers, [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], headers)
        self.check_html([[0,1,2],[3,4,5],[6,7,8]], headers, [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], headers)

        # check with dict and headers='keys'
        tdict = {
            'Col0': [0,3,6],
            'Col1': [1,4,7],
            'Col2': [2,5,8],
        }
        self.check_html(tdict, 'keys', [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], headers)

        # check list of dicts
        dlist = [
            {'Col0': 0, 'Col1': 1, 'Col2': 2},
            {'Col0': 3, 'Col1': 4, 'Col2': 5},
            {'Col0': 6, 'Col1': 7, 'Col2': 8},
            ]
        self.check_html(dlist, 'keys', [['0', '1', '2'], ['3', '4', '5'], ['6', '7', '8']], headers)

    def test_missing_vals(self):
        headers = ['Strings', 'Floats', 'Something else']
        expected = """
| ----------------------------------------------- |
| Strings               | Floats | Something else |
| --------------------- | ------ | -------------- |
| foobar blah           |    1.0 | N/A            |
| asdfas dffff          |  3.142 | N/A            |
| hello world blah blah |   9.87 | N/A            |
| ----------------------------------------------- |
"""
        self.check_text('text', self.table_row_iter('str', 'float', None), headers, expected, missing_val='N/A')

    def test_word_wrap(self):
        headers = ['Strings', 'Floats', 'Something else']
        expected = """
| ---------------------------------- |
| Strings     | Floats | Something   |
|             |        | else        |
| ----------- | ------ | ----------- |
| foobar blah |    1.0 | N/A         |
| asdfas      |  3.142 | N/A         |
| dffff       |        |             |
| hello world |   9.87 | N/A         |
| blah blah   |        |             |
| ---------------------------------- |
"""
        self.check_text('text', self.table_row_iter('str', 'float', None), headers, expected, missing_val='N/A', max_width=38)

    def test_word_wrap_box_grid(self):
        headers = ['Strings', 'Floats', 'Something else']
        expected = """
╔═════════════╤════════╤═════════════╗
║ Strings     ┊ Floats ┊ Something   ║
║             ┊        ┊ else        ║
╠═════════════╪════════╪═════════════╣
║ foobar blah ┊    1.0 ┊ N/A         ║
╟┈┈┈┈┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ asdfas      ┊  3.142 ┊ N/A         ║
║ dffff       ┊        ┊             ║
╟┈┈┈┈┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ hello world ┊   9.87 ┊ N/A         ║
║ blah blah   ┊        ┊             ║
╚═════════════╧════════╧═════════════╝
"""
        self.check_text('box_grid', self.table_row_iter('str', 'float', None), headers, expected, missing_val='N/A', max_width=38)

    def test_align(self):
        headers = ['Strings', 'Floats', 'Something else']
        column_meta = [{'header_align': 'center'}, {}, {'align': 'center', 'header_align': 'center'}]
        expected = """
| ---------------------------------- |
|   Strings   | Floats |  Something  |
|             |        |    else     |
| ----------- | ------ | ----------- |
| foobar blah |    1.0 |     N/A     |
| asdfas      |  3.142 |     N/A     |
| dffff       |        |             |
| hello world |   9.87 |     N/A     |
| blah blah   |        |             |
| ---------------------------------- |
"""
        self.check_text('text', self.table_row_iter('str', 'float', None), headers, expected,
                        column_meta=column_meta, missing_val='N/A', max_width=38)

    def test_basic_tabulator(self):
        # for now, just check that it doesn't crash
        headers = ['Strings', 'Floats', 'Something else']
        table = generate_table(self.table_row_iter('str', 'float', None),
                               tablefmt='tabulator', headers=headers)
        tstr = str(table)

    def test_embedded_newline(self):
        cells = [
            ["#", "title A", "title B"],
            ["1", "lorem", "ipsum dolor sit amet"],
            ["2", "lorem", "ipsum\ndolor sit amet"],
            ["3", "lorem", "ipsum dolor sit amet"],
        ]
        expected = """
╔═══╤═════════╤══════════════════════╗
║ # ┊ title A ┊ title B              ║
╠═══╪═════════╪══════════════════════╣
║ 1 ┊ lorem   ┊ ipsum dolor sit amet ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 2 ┊ lorem   ┊ ipsum                ║
║   ┊         ┊ dolor sit amet       ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 3 ┊ lorem   ┊ ipsum dolor sit amet ║
╚═══╧═════════╧══════════════════════╝
"""
        self.check_text('box_grid', cells, 'firstrow', expected)

    def test_embedded_newline2(self):
        cells = [
            ["#", "title A", "title B"],
            ["1", "lorem", "ipsum dolor sit amet"],
            ["2", "lorem", "ipsum\ndolor sit\namet"],
            ["3", "lorem", "ipsum dolor sit amet"],
        ]
        expected = """
╔═══╤═════════╤══════════════════════╗
║ # ┊ title A ┊ title B              ║
╠═══╪═════════╪══════════════════════╣
║ 1 ┊ lorem   ┊ ipsum dolor sit amet ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 2 ┊ lorem   ┊ ipsum                ║
║   ┊         ┊ dolor sit            ║
║   ┊         ┊ amet                 ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 3 ┊ lorem   ┊ ipsum dolor sit amet ║
╚═══╧═════════╧══════════════════════╝
"""
        self.check_text('box_grid', cells, 'firstrow', expected)

    def test_non_string_header(self):
        cells = [
            ["#", "title A", "1.0"],
            ["1", "lorem", "ipsum dolor sit amet"],
            ["2", "lorem", "ipsum dolor sit amet"],
            ["3", "lorem", "ipsum dolor sit amet"],
        ]
        expected = """
╔═══╤═════════╤══════════════════════╗
║ # ┊ title A ┊ 1.0                  ║
╠═══╪═════════╪══════════════════════╣
║ 1 ┊ lorem   ┊ ipsum dolor sit amet ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 2 ┊ lorem   ┊ ipsum dolor sit amet ║
╟┈┈┈┿┈┈┈┈┈┈┈┈┈┿┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈╢
║ 3 ┊ lorem   ┊ ipsum dolor sit amet ║
╚═══╧═════════╧══════════════════════╝
"""
        self.check_text('box_grid', cells, 'firstrow', expected)



"""Define the ExplicitFuncComp class."""

import sys
import traceback
import numpy as np

from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.core.constants import INT_DTYPE
import openmdao.func_api as omf
from openmdao.components.func_comp_common import _check_var_name, _copy_with_ignore, _add_options, \
    jac_forward, jac_reverse, jacvec_prod, _get_tangents
from openmdao.utils.array_utils import shape_to_len

try:
    import jax
    from jax import jit
    import jax.numpy as jnp
    from jax.config import config
    config.update("jax_enable_x64", True)  # jax by default uses 32 bit floats
except Exception:
    _, err, tb = sys.exc_info()
    if not isinstance(err, ImportError):
        traceback.print_tb(tb)
    jax = None

if jax is not None:
    try:
        from jax import Array as JaxArray
    except ImportError:
        # versions of jax before 0.3.18 do not have the jax.Array base class
        raise RuntimeError(f"An unsupported version of jax is installed. "
                           "OpenMDAO requires 'jax>=4.0' and 'jaxlib>=4.0'. "
                           "Try 'pip install openmdao[jax]' with Python>=3.8.")


class ExplicitFuncComp(ExplicitComponent):
    """
    A component that wraps a python function.

    Parameters
    ----------
    compute : function
        The function to be wrapped by this Component.
    compute_partials : function or None
        If not None, call this function when computing partials.
    **kwargs : named args
        Args passed down to ExplicitComponent.

    Attributes
    ----------
    _compute : callable
        The function wrapper used by this component.
    _compute_jax : callable
        Function decorated to ensure use of jax numpy.
    _compute_partials : function or None
        If not None, call this function when computing partials.
    _tangents : tuple
        Tuple of parts of the tangent matrix cached for jax derivative computation.
    """

    def __init__(self, compute, compute_partials=None, **kwargs):
        """
        Initialize attributes.
        """
        super().__init__(**kwargs)
        self._compute = omf.wrap(compute)
        # in case we're doing jit, force setup of wrapped func because we compute output shapes
        # during setup and that won't work on a jit compiled function
        if self._compute._call_setup:
            self._compute._setup()

        if self._compute._use_jax:
            self.options['use_jax'] = True

        if self.options['use_jax']:
            if jax is None:
                raise RuntimeError(f"{self.msginfo}: jax is not installed. "
                                   "Try 'pip install openmdao[jax]' with Python>=3.8.")
            self._compute_jax = omf.jax_decorate(self._compute._f)

        self._tangents = None

        self._compute_partials = compute_partials
        if self.options['use_jax'] and self.options['use_jit']:
            static_argnums = [i for i, m in enumerate(self._compute._inputs.values())
                              if 'is_option' in m]
            try:
                self._compute_jax = jit(self._compute_jax, static_argnums=static_argnums)
            except Exception as err:
                raise RuntimeError(f"{self.msginfo}: failed jit compile of compute function: {err}")

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        super()._declare_options()
        _add_options(self)

    def setup(self):
        """
        Define out inputs and outputs.
        """
        optignore = {'is_option'}
        use_jax = self.options['use_jax'] and jax is not None

        for name, meta in self._compute.get_input_meta():
            _check_var_name(self, name)
            if 'is_option' in meta and meta['is_option']:
                kwargs = _copy_with_ignore(meta, omf._allowed_declare_options_args,
                                           ignore=optignore)
                self.options.declare(name, **kwargs)
            else:
                kwargs = omf._filter_dict(meta, omf._allowed_add_input_args)
                if use_jax:
                    # make sure internal openmdao values are numpy arrays and not jax Arrays
                    self._dev_arrays_to_np_arrays(kwargs)
                self.add_input(name, **kwargs)

        for i, (name, meta) in enumerate(self._compute.get_output_meta()):
            _check_var_name(self, name)
            kwargs = _copy_with_ignore(meta, omf._allowed_add_output_args, ignore=('resid',))
            if use_jax:
                # make sure internal openmdao values are numpy arrays and not jax Arrays
                self._dev_arrays_to_np_arrays(kwargs)
            self.add_output(name, **kwargs)

    def _dev_arrays_to_np_arrays(self, meta):
        if 'val' in meta:
            if isinstance(meta['val'], JaxArray):
                meta['val'] = np.asarray(meta['val'])

    def _linearize(self, jac=None, sub_do_ln=False):
        """
        Compute jacobian / factorization. The model is assumed to be in a scaled state.

        Parameters
        ----------
        jac : Jacobian or None
            Ignored.
        sub_do_ln : bool
            Flag indicating if the children should call linearize on their linear solvers.
        """
        if self.options['use_jax']:
            self._check_first_linearize()
            self._jax_linearize()
        else:
            super()._linearize(jac, sub_do_ln)

    def _jax_linearize(self):
        """
        Compute the jacobian using jax.

        This updates self._jacobian.
        """
        inames = list(self._compute.get_input_names())
        # argnums specifies which position args are to be differentiated
        argnums = [i for i, m in enumerate(self._compute._inputs.values()) if 'is_option' not in m]
        # keep this around for use locally even if we pass None as argnums to jax
        argidxs = argnums
        if len(argnums) == len(inames):
            argnums = None  # speedup if there are no static args
        osize = len(self._outputs)
        isize = len(self._inputs)
        invals = list(self._func_values(self._inputs))
        coloring = self._coloring_info['coloring']
        func = self._compute_jax

        if self._mode == 'rev':  # use reverse mode to compute derivs
            outvals = tuple(self._outputs.values())
            tangents = self._get_tangents(outvals, 'rev', coloring)
            if coloring is None:
                j = np.empty((osize, isize), dtype=float)
                cstart = cend = 0
                for i, a in zip(argidxs, jac_reverse(func, argnums, tangents)(*invals)):
                    if isinstance(invals[i], np.ndarray):
                        cend += invals[i].size
                    else:  # must be a scalar
                        cend += 1
                    a = np.asarray(a)
                    if a.ndim < 2:
                        j[:, cstart:cend] = a.reshape((a.size, 1))
                    else:
                        j[:, cstart:cend] = a.reshape((a.shape[0], cend - cstart))
                    cstart = cend
            else:
                j = [np.asarray(a).reshape((a.shape[0], shape_to_len(a.shape[1:])))
                     for a in jac_reverse(func, argnums, tangents)(*invals)]
                j = coloring.expand_jac(np.hstack(j), 'rev')
        else:
            tangents = self._get_tangents(invals, 'fwd', coloring, argnums)
            if coloring is None:
                j = np.empty((osize, isize), dtype=float)
                start = end = 0
                for a in jac_forward(func, argnums, tangents)(*invals):
                    a = np.asarray(a)
                    if a.ndim < 2:
                        a = a.reshape((1, a.size))
                    else:
                        a = a.reshape((shape_to_len(a.shape[:-1]), a.shape[-1]))
                    end += a.shape[0]
                    if osize == 1:
                        j[0, start:end] = a
                    else:
                        j[start:end, :] = a
                    start = end
            else:
                j = [np.asarray(a).reshape((shape_to_len(a.shape[:-1]), a.shape[-1]))
                     for a in jac_forward(func, argnums, tangents)(*invals)]
                j = coloring.expand_jac(np.vstack(j), 'fwd')

        self._jacobian.set_dense_jac(self, j)

    def _get_tangents(self, vals, direction, coloring=None, argnums=None):
        """
        Return a tuple of tangents values for use with vmap.

        Parameters
        ----------
        vals : list
            List of function input values.
        direction : str
            Derivative computation direction ('fwd' or 'rev').
        coloring : Coloring or None
            If not None, the Coloring object used to compute a compressed tangent array.
        argnums : list of int or None
            Indices of dynamic (differentiable) function args.

        Returns
        -------
        tuple of ndarray or ndarray
            The tangents values to be passed to vmap.
        """
        if self._tangents is None:
            self._tangents = _get_tangents(vals, direction, coloring, argnums)
        return self._tangents

    def compute(self, inputs, outputs):
        """
        Compute the result of calling our function with the given inputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables.
        outputs : Vector
            Unscaled, dimensional output variables.
        """
        outputs.set_vals(self._compute(*self._func_values(inputs)))

    def declare_partials(self, *args, **kwargs):
        """
        Declare information about this component's subjacobians.

        Parameters
        ----------
        *args : list
            Positional args to be passed to base class version of declare_partials.
        **kwargs : dict
            Keyword args  to be passed to base class version of declare_partials.

        Returns
        -------
        dict
            Metadata dict for the specified partial(s).
        """
        return super().declare_partials(*args, **kwargs)

    def _setup_partials(self):
        """
        Check that all partials are declared.
        """
        for kwargs in self._compute.get_declare_partials():
            self.declare_partials(**kwargs)

        kwargs = self._compute.get_declare_coloring()
        if kwargs is not None:
            self.declare_coloring(**kwargs)

        super()._setup_partials()

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Sub-jac components written to partials[output_name, input_name].
        """
        if self._compute_partials is None:
            return

        self._compute_partials(*self._func_values(inputs), partials)

    def _func_values(self, inputs):
        """
        Yield current function input args.

        Parameters
        ----------
        inputs : Vector
            The input vector.

        Yields
        ------
        object
            Value of current function input variable.
        """
        inps = inputs.values()

        for name, meta in self._compute._inputs.items():
            if 'is_option' in meta:
                yield self.options[name]
            else:
                yield next(inps)

    def _compute_coloring(self, recurse=False, **overrides):
        """
        Compute a coloring of the partial jacobian.

        This assumes that the current System is in a proper state for computing derivatives.
        It just calls the base class version and then resets the tangents so that after coloring
        a new set of compressed tangents values can be computed.

        Parameters
        ----------
        recurse : bool
            If True, recurse from this system down the system hierarchy.  Whenever a group
            is encountered that has specified its coloring metadata, we don't recurse below
            that group unless that group has a subsystem that has a nonlinear solver that uses
            gradients.
        **overrides : dict
            Any args that will override either default coloring settings or coloring settings
            resulting from an earlier call to declare_coloring.

        Returns
        -------
        list of Coloring
            The computed colorings.
        """
        ret = super()._compute_coloring(recurse, **overrides)
        self._tangents = None  # reset to compute new colored tangents later
        return ret



"""
Utilities for working with files.
"""

import sys
import os
import importlib
import types
from fnmatch import fnmatch
from os.path import join, basename, dirname, isfile, split, splitext, abspath


def get_module_path(fpath):
    """
    Given a module filename, return its full Python module path.

    This includes enclosing packages and is based on existence of ``__init__.py`` files.

    Parameters
    ----------
    fpath : str
        Pathname of file.

    Returns
    -------
    str or None
        Full module path of the given file.  Returns None if the file is not part of a package.
    """
    fpath = abspath(fpath)
    if basename(fpath).startswith('__init__.'):
        pnames = []
    else:
        pnames = [splitext(basename(fpath))[0]]
    path = dirname(fpath)

    initfile = join(path, '__init__.py')
    if not isfile(initfile):
        return None

    while isfile(initfile):
        path, pname = split(path)
        pnames.append(pname)
        initfile = join(path, '__init__.py')

    return '.'.join(pnames[::-1])


def package_iter(start_dir='.', dir_includes=None, dir_excludes=(), file_includes=None,
                 file_excludes=()):
    """
    Iterate over python files in packages (recursively) starting in start_dir.

    NOTE : all *_includes and *_excludes are applied to *local* directory and file names.

    Parameters
    ----------
    start_dir : str
        Starting directory.
    dir_includes : iter of str or None
        Glob patterns for directory inclusion. Be careful here because dir names are local,
        so, for example, if includes=('foo',), then directory 'foo' would be included, but
        any subdirectories of 'foo' that were not also named 'foo' would not.
    dir_excludes : iter of str
        Glob patterns for directory exclusion.
    file_includes : iter of str or None
        Glob patterns for file inclusion.
    file_excludes : iter of str
        Glob patterns for file exclusion.

    Yields
    ------
    str
        Filenames (full path from start_dir).
    """
    file_includes = set() if file_includes is None else set(file_includes)
    file_includes.add('*.py')
    file_excludes = set() if file_excludes is None else set(file_excludes)
    file_excludes.update(('_*', 'test_*', 'api.py', 'parallel_api.py'))
    dir_excludes = set() if dir_excludes is None else set(dir_excludes)
    dir_excludes.update(('test', 'tests'))

    for f in files_iter(start_dir, dir_includes=dir_includes, dir_excludes=dir_excludes,
                        file_includes=file_includes, file_excludes=file_excludes,
                        package_only=True):
        yield f


def files_iter(start_dir='.', dir_includes=None, dir_excludes=(),
               file_includes=None, file_excludes=(), package_only=False):
    """
    Iterate over files (recursively) starting in start_dir.

    NOTE : all *_includes and *_excludes are applied to *local* directory and file names.

    Parameters
    ----------
    start_dir : str
        Starting directory.
    dir_includes : iter of str or None
        Glob patterns for directory inclusion. Be careful here because dir names are local,
        so, for example, if includes=('foo',), then directory 'foo' would be included, but
        any subdirectories of 'foo' that were not also named 'foo' would not.
    dir_excludes : iter of str
        Glob patterns for directory exclusion.
    file_includes : iter of str or None
        Glob patterns for file inclusion.
    file_excludes : iter of str
        Glob patterns for file exclusion.
    package_only : bool
        If True, only yield files that are contained in a python package.

    Yields
    ------
    str
        Filenames (full path from start_dir).
    """
    for root, dirs, files in os.walk(start_dir):
        if package_only and '__init__.py' not in files:
            dirs[:] = []
            continue
        for pat in dir_excludes:
            dirs[:] = sorted([d for d in dirs if not fnmatch(d, pat)])
        if dir_includes:
            incdirs = set()
            for pat in dir_includes:
                incdirs.update(d for d in dirs if fnmatch(d, pat))
            dirs[:] = sorted(incdirs)
        for f in files:
            for pat in file_excludes:
                if fnmatch(f, pat):
                    break
            else:
                if file_includes:
                    for pat in file_includes:
                        if fnmatch(f, pat):
                            yield join(root, f)
                else:
                    yield join(root, f)


def _to_filename(spec):
    """
    Return the filename part of the given testspec or the full string if the string is a filename.

    Parameters
    ----------
    spec : str
        The filename or testspec.

    Returns
    -------
    str
        The filename.
    """
    if ':' in spec and not os.path.isfile(spec):
        fname, _ = spec.rsplit(':', 1)
        if not fname.endswith('.py'):
            try:
                mod = importlib.import_module(fname)
                return mod.__file__
            except ImportError:
                return spec
        return fname

    return spec


def _load_and_exec(script_name, user_args):
    """
    Load and exec the given script as __main__.

    Parameters
    ----------
    script_name : str
        The name of the script to load and exec.
    user_args : list of str
        Args to be passed to the user script.
    """
    if ':' in script_name and not os.path.isfile(script_name):
        return _load_and_run_test(script_name)

    sys.path.insert(0, os.path.dirname(script_name))

    sys.argv[:] = [script_name] + user_args

    with open(script_name, 'rb') as fp:
        code = compile(fp.read(), script_name, 'exec')

    globals_dict = {
        '__file__': script_name,
        '__name__': '__main__',
        '__package__': None,
        '__cached__': None,
    }

    exec(code, globals_dict)  # nosec: private, internal use only


def fname2mod_name(fname):
    """
    Convert a string to a valid python module name.

    Parameters
    ----------
    fname : str
        The filename to convert.

    Returns
    -------
    str
        A valid module name corresponding to the given filename.
    """
    to_replace = ['-', ' ', '.', '(', ')', '[', ']', '{', '}', '=', '+'
                  '!', '@', '#', '$', '%', '^', '&', '*', '~', '`',
                  ';', ':', '"', "'", '<', '>', '?', '/', '\\', '|']

    if not fname.endswith('.py'):
        raise ValueError(f"'{s}' does not end with '.py'")

    s = os.path.basename(fname).rsplit('.', 1)[0]

    for c in to_replace:
        s = s.replace(c, '_')
    return s


def _load_and_run_test(testspec):
    """
    Load and run an individual test function.

    Parameters
    ----------
    testspec : str
        <fpath_or_modpath>:<testcase>.<method> OR <fpath_or_modpath>:<function>
    """
    syspath_save = sys.path[:]

    modpath, funcpath = testspec.rsplit(':', 1)
    orig_modpath = modpath

    if modpath.endswith('.py'):
        modpath = get_module_path(modpath)
        if modpath is None:
            # create a module dynamically
            modpath = fname2mod_name(orig_modpath)
            mod = types.ModuleType(modpath)
            sys.modules[modpath] = mod
            mod.__file__ = modpath
            mod.__name__ = modpath
            mod.__package__ = None
            mod.__cached__ = None
            with open(orig_modpath, 'rb') as fp:
                code = compile(fp.read(), orig_modpath, 'exec')
            exec(code, mod.__dict__)  # nosec: private, internal use only
        else:
            mod = importlib.import_module(modpath)
    else:
        mod = importlib.import_module(modpath)

    try:
        return _run_test_func(mod, funcpath)
    finally:
        sys.path = syspath_save


def _run_test_func(mod, funcpath):
    """
    Run the given TestCase method or test function in the given module.

    Parameters
    ----------
    mod : module
        The module where the test resides.
    funcpath : str
        Either <testcase>.<method_name> or <func_name>.

    Returns
    -------
    object
        In the case of a module level function call, returns whatever the function returns.
    """
    parts = funcpath.split('.', 1)
    if len(parts) == 2:
        tcase_name, method_name = parts
        testcase = getattr(mod, tcase_name)(methodName=method_name)
        setup = getattr(testcase, 'setUp', None)
        if setup is not None:
            setup()
        getattr(testcase, method_name)()
        teardown = getattr(testcase, 'tearDown', None)
        if teardown:
            teardown()
    else:
        funcname = parts[0]
        return getattr(mod, funcname)()


if sys.version_info >= (3, 8):
    from importlib.metadata import entry_points

    if sys.version_info >= (3, 10):
        def _eps_get(group):
            eps = entry_points().select(group=group)
            for name in eps.names:
                yield eps[name]
    else:
        def _eps_get(group):
            eps = entry_points()
            if group in eps:
                yield from eps[group]

    def _iter_entry_points(group):
        # there seems to be a bug currently where entry points can show up more than
        # once in the iterator, so keep track of the ones we've already seen.
        # TODO: revisit later to see if we can remove the check
        seen = set()
        for ep in _eps_get(group):
            if ep.name not in seen:
                seen.add(ep.name)
                yield ep
else:
    try:
        import pkg_resources
    except ImportError:
        def _iter_entry_points(group):
            issue_warning("Can't retrieve entry points because pkg_resources is not installed. "
                          "Either install it using 'pip install setuptools' or upgrade to python "
                          "3.8 or newer.")
            return ()
    else:
        def _iter_entry_points(group):
            yield from pkg_resources.iter_entry_points(group)


def text2html(text, title='', style=None):
    """
    Wrap the given text for display as an html file.

    Returns an html syntax string that can be written to a file.

    Parameters
    ----------
    text : str
        Text to be displayed.
    title : str
        Title to display above text.
    style : str or None
        If not None, use as the contents of the style block for the enclosing <pre> tag.

    Returns
    -------
    str
        Content string to create an html file.
    """
    if style is None:
        style = """
            display: block;
            font-family: monospace;
            font-size: 1.5em;
            white-space: pre;
            margin: 1em 0;
        """
    return """
<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        .center {
            display: block;
            margin-left: auto;
            margin-right: auto;
            width: 90%;
        }
        h2 {text-align: center;}
        pre {""" + style + """
        }
    </style>
</head>
<body>
<h2>""" + title + """</h2>
<pre>
""" + text + """
</pre>
</body>
</html>
"""


def image2html(imagefile, title='', alt=''):
    """
    Wrap the given image for display as an html file.

    Returns an html syntax string that can be written to a file.

    Parameters
    ----------
    imagefile : str
        Name of image file to be displayed.
    title : str
        The page title.
    alt : str
        Set the alt text for the image.

    Returns
    -------
    str
        Content string to create an html file.
    """
    return """
<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        h2 {text-align: center;}
        .center {
            display: block;
            margin-left: auto;
            margin-right: auto;
            width: 80%;
        }
    </style>
</head>
<body>
<h2>""" + title + "</h2>" + f"""
<img src="{imagefile}" alt="{alt}" class="center"></img>

</body>
</html>
"""



import unittest
import numpy as np

import openmdao.api as om
import openmdao.core.problem
import openmdao.utils.hooks as hooks
from openmdao.utils.assert_utils import assert_warning


def make_hook(name):
    def hook_func(obj):
        obj.calls.append(name)
    return hook_func


def hooks_active(f):
    # turn hooks on and off around a hooks test
    def _wrapper(*args, **kwargs):
        hooks._reset_all_hooks()
        hooks.use_hooks = True
        try:
            f(*args, **kwargs)
        finally:
            hooks.use_hooks = False
    return _wrapper


class HooksTestCase(unittest.TestCase):
    def setUp(self):
        openmdao.core.problem._clear_problem_names()  # need to reset these to simulate separate runs

    def build_model(self, name='problem1'):
        prob = om.Problem(name=name)
        prob.calls = []
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 3.0))
        model.add_subsystem('p2', om.IndepVarComp('y', -4.0))
        model.add_subsystem('comp', om.ExecComp("f_xy=2.0*x+3.0*y"))

        model.connect('p1.x', 'comp.x')
        model.connect('p2.y', 'comp.y')
        prob.setup()
        return prob

    @hooks_active
    def test_ncalls(self):
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final'), post=make_hook('post_final'), ncalls=2)
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final2'), post=make_hook('post_final2'))

        prob = self.build_model()
        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['pre_final', 'pre_final2', 'post_final', 'post_final2',
                                      'pre_final', 'pre_final2', 'post_final', 'post_final2',
                                      'pre_final2', 'post_final2',
                                     ])

    @hooks_active
    def test_exit(self):
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final'), post=make_hook('post_final'))
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final2'), post=make_hook('post_final2'), exit=True)

        prob = self.build_model()
        try:
            prob.run_model()
            prob.run_model()
            prob.run_model()
        except SystemExit:
            self.assertEqual(prob.calls, ['pre_final', 'pre_final2', 'post_final', 'post_final2'])
        else:
            self.fail("sys.exit() was not called")

    @hooks_active
    def test_multiwrap(self):
        pre_final = make_hook('pre_final')
        post_final = make_hook('post_final')
        hooks._register_hook('final_setup', 'Problem', pre=pre_final, post=post_final)
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final2'), post=make_hook('post_final2'))

        prob = self.build_model()
        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['pre_final', 'pre_final2', 'post_final', 'post_final2',
                                      'pre_final', 'pre_final2', 'post_final', 'post_final2',
                                      'pre_final', 'pre_final2', 'post_final', 'post_final2',
                                     ])

        hooks._unregister_hook('final_setup', 'Problem', pre=pre_final, post=False)
        prob.calls = []

        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['pre_final2', 'post_final', 'post_final2',
                                      'pre_final2', 'post_final', 'post_final2',
                                      'pre_final2', 'post_final', 'post_final2',
                                     ])

        hooks._unregister_hook('final_setup', 'Problem', pre=True, post=False)
        prob.calls = []

        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['post_final', 'post_final2',
                                      'post_final', 'post_final2',
                                      'post_final', 'post_final2',
                                     ])

    @hooks_active
    def test_multiwrap_mixed_None_inst(self):
        pre_final = make_hook('pre_final')
        post_final = make_hook('post_final')
        hooks._register_hook('final_setup', 'Problem', pre=pre_final, post=post_final)
        hooks._register_hook('final_setup', 'Problem', inst_id='problem1', ncalls=2,
                             pre=make_hook('pre_final1'), post=make_hook('post_final1'))
        hooks._register_hook('final_setup', 'Problem', inst_id='problem2',
                             pre=make_hook('pre_final2'), post=make_hook('post_final2'))

        probs = [self.build_model(f"problem{i+1}") for i in range(2)]
        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['pre_final', 'pre_final1', 'post_final', 'post_final1', 'pre_final', 'pre_final1', 'post_final', 'post_final1', 'pre_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['pre_final', 'pre_final2', 'post_final', 'post_final2', 'pre_final', 'pre_final2', 'post_final', 'post_final2', 'pre_final', 'pre_final2', 'post_final', 'post_final2'])

        hooks._unregister_hook('final_setup', 'Problem', pre=pre_final, post=False)

        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['post_final', 'post_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['pre_final2', 'post_final', 'post_final2', 'pre_final2', 'post_final', 'post_final2', 'pre_final2', 'post_final', 'post_final2'])

        hooks._unregister_hook('final_setup', 'Problem', pre=True, post=False)

        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['post_final', 'post_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['post_final', 'post_final2', 'post_final', 'post_final2', 'post_final', 'post_final2'])

    @hooks_active
    def test_multiwrap_mixed_inst_None(self):
        pre_final = make_hook('pre_final')
        post_final = make_hook('post_final')
        hooks._register_hook('final_setup', 'Problem', inst_id='problem1', ncalls=2,
                             pre=make_hook('pre_final1'), post=make_hook('post_final1'))
        hooks._register_hook('final_setup', 'Problem', inst_id='problem2',
                             pre=make_hook('pre_final2'), post=make_hook('post_final2'))
        hooks._register_hook('final_setup', 'Problem', pre=pre_final, post=post_final)

        probs = [self.build_model(f"problem{i+1}") for i in range(2)]
        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['pre_final1', 'pre_final', 'post_final1', 'post_final', 'pre_final1', 'pre_final', 'post_final1', 'post_final', 'pre_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['pre_final2', 'pre_final', 'post_final2', 'post_final', 'pre_final2', 'pre_final', 'post_final2', 'post_final', 'pre_final2', 'pre_final', 'post_final2', 'post_final'])

        hooks._unregister_hook('final_setup', 'Problem', pre=pre_final, post=False)

        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['post_final', 'post_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['pre_final2', 'post_final2', 'post_final', 'pre_final2', 'post_final2', 'post_final', 'pre_final2', 'post_final2', 'post_final'])

        hooks._unregister_hook('final_setup', 'Problem', pre=True, post=False)

        for prob in probs:
            prob.calls = []
            for i in range(3):
                prob.run_model()

        self.assertEqual(probs[0].calls, ['post_final', 'post_final', 'post_final'])
        self.assertEqual(probs[1].calls, ['post_final2', 'post_final', 'post_final2', 'post_final', 'post_final2', 'post_final'])

    @hooks_active
    def test_problem_hooks(self):
        hooks._register_hook('setup', 'Problem', pre=make_hook('pre_setup'), post=make_hook('post_setup'))
        hooks._register_hook('final_setup', 'Problem', pre=make_hook('pre_final'), post=make_hook('post_final'))
        hooks._register_hook('run_model', 'Problem', pre=make_hook('pre_run_model'), post=make_hook('post_run_model'))

        prob = self.build_model()

        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['pre_setup', 'post_setup',
                                        'pre_run_model', 'pre_final', 'post_final', 'post_run_model',
                                        'pre_run_model', 'pre_final', 'post_final', 'post_run_model',
                                        'pre_run_model', 'pre_final', 'post_final', 'post_run_model',
                                        ])

        np.testing.assert_allclose(prob['comp.f_xy'], -6.0)

        hooks._unregister_hook('setup', 'Problem', pre=False)
        hooks._unregister_hook('final_setup', 'Problem')
        hooks._unregister_hook('run_model', 'Problem', post=False)
        prob.calls = []

        prob.setup()
        prob.run_model()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, ['pre_setup', 'post_run_model', 'post_run_model', 'post_run_model'])

        hooks._unregister_hook('setup', 'Problem')

        msg = "No hook found for method 'final_setup' for class 'Problem' and instance 'None'."

        # already removed final_setup hooks earlier, so expect a warning here
        with assert_warning(UserWarning, msg):
            hooks._unregister_hook('final_setup', 'Problem')

        hooks._unregister_hook('run_model', 'Problem')
        prob.calls = []

        prob.setup()
        prob.run_model()
        prob.run_model()

        self.assertEqual(prob.calls, [])
        self.assertEqual(len(hooks._hooks), 0)  # should be no hooks left

    @hooks_active
    def test_problem_hooks_kwargs(self):

        x0 = 33.0
        y0 = 44.0

        def set_prob_vars_hook_func(prob, **kwargs):
            if 'x0' in kwargs:
                prob['p1.x'] = kwargs['x0']
            if 'y0' in kwargs:
                prob['p2.y'] = kwargs['y0']

        hooks._register_hook('final_setup', 'Problem', pre=set_prob_vars_hook_func, x0=x0, y0=y0)

        prob = self.build_model()

        prob.run_model()

        self.assertEqual(prob['comp.x'], x0)
        self.assertEqual(prob['comp.y'], y0)


if __name__ == '__main__':
    unittest.main()



"""Test the inputs report. """
import unittest
import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.double_sellar import DoubleSellar
from openmdao.utils.assert_utils import assert_near_equal, assert_warning
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.visualization.inputs_report.inputs_report import inputs_report

from openmdao.utils.mpi import MPI

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None

@use_tempdirs
class TestInputsReport(unittest.TestCase):

    def test_inputs_reports(self):
        prob = om.Problem(model=DoubleSellar(), reports=None)
        model = prob.model

        g1 = model.g1
        g1.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g1.nonlinear_solver.options['rtol'] = 1.0e-5
        g1.linear_solver = om.DirectSolver(assemble_jac=True)
        g1.options['assembled_jac_type'] = 'dense'

        g2 = model.g2
        g2.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        g2.nonlinear_solver.options['rtol'] = 1.0e-5
        g2.linear_solver = om.DirectSolver(assemble_jac=True)
        g2.options['assembled_jac_type'] = 'dense'

        model.nonlinear_solver = om.NewtonSolver()
        model.linear_solver = om.ScipyKrylov(assemble_jac=True)
        model.options['assembled_jac_type'] = 'dense'

        model.nonlinear_solver.options['solve_subsystems'] = True

        prob.setup()
        prob.run_model()

        assert_near_equal(prob['g1.y1'], 0.64, .00001)
        assert_near_equal(prob['g1.y2'], 0.80, .00001)
        assert_near_equal(prob['g2.y1'], 0.64, .00001)
        assert_near_equal(prob['g2.y2'], 0.80, .00001)

        expected = """| Absolute Name | Input Name | Source Name  | Source is IVC | Source is DV | Units | Shape | Tags | Val    | Min Val | Max Val | Absolute Source |
| :------------ | :--------- | :----------- | :-----------: | :----------: | :---- | :---- | :--- | :----- | ------: | ------: | :-------------- |
| g1.d1.z       | g1.z       | _auto_ivc.v0 |     True      |    False     |       | (2,)  | []   | [0 0]  |       0 |       0 | _auto_ivc.v0    |
| g1.d1.x       | g1.x       | g2.y2        |     False     |    False     |       | (1,)  | []   | [0.8]  |     0.8 |     0.8 | g2.d2.y2        |
| g1.d1.y2      | g1.y2      | g1.y2        |     False     |    False     |       | (1,)  | []   | [0.8]  |     0.8 |     0.8 | g1.d2.y2        |
| g1.d2.z       | g1.z       | _auto_ivc.v0 |     True      |    False     |       | (2,)  | []   | [0 0]  |       0 |       0 | _auto_ivc.v0    |
| g1.d2.y1      | g1.y1      | g1.y1        |     False     |    False     |       | (1,)  | []   | [0.64] |    0.64 |    0.64 | g1.d1.y1        |
| g2.d1.z       | g2.z       | _auto_ivc.v1 |     True      |    False     |       | (2,)  | []   | [0 0]  |       0 |       0 | _auto_ivc.v1    |
| g2.d1.x       | g2.x       | g1.y2        |     False     |    False     |       | (1,)  | []   | [0.8]  |     0.8 |     0.8 | g1.d2.y2        |
| g2.d1.y2      | g2.y2      | g2.y2        |     False     |    False     |       | (1,)  | []   | [0.8]  |     0.8 |     0.8 | g2.d2.y2        |
| g2.d2.z       | g2.z       | _auto_ivc.v1 |     True      |    False     |       | (2,)  | []   | [0 0]  |       0 |       0 | _auto_ivc.v1    |
| g2.d2.y1      | g2.y1      | g2.y1        |     False     |    False     |       | (1,)  | []   | [0.64] |    0.64 |    0.64 | g2.d1.y1        |
"""
        inputs_report(prob, outfile='temp_inputs_report.md', display=True, precision=6, title=None, tablefmt='github')
        with open('temp_inputs_report.md') as f:
            report_content = f.read()
        self.assertEqual(expected, report_content)

    def test_zero_size_input(self):
        class TestComp(om.ExplicitComponent):
            def setup(self):
                self.add_input('foo', shape=0)
                self.add_output('bar', shape=0)

            def compute(self, inputs, outputs):
                outputs['bar'] = inputs['foo']

        p = om.Problem()
        p.model.add_subsystem('comp', TestComp())
        p.setup()

        p.set_val('comp.foo', np.array([]))
        p.run_model()

        expected = """| Absolute Name | Input Name | Source Name  | Source is IVC | Source is DV | Units | Shape | Tags | Val | Min Val | Max Val | Absolute Source |
| :------------ | :--------- | :----------- | :-----------: | :----------: | :---- | :---- | :--- | :-- | :------ | :------ | :-------------- |
| comp.foo      | comp.foo   | _auto_ivc.v0 |     True      |    False     |       | (0,)  | []   | []  | []      | []      | _auto_ivc.v0    |
"""

        inputs_report(p, outfile='temp_inputs_report.md', display=True, precision=6, title=None, tablefmt='github')
        with open('temp_inputs_report.md') as f:
            report_content = f.read()
        self.assertEqual(expected, report_content)


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
@use_tempdirs
class TestInputReportsMPI(unittest.TestCase):
    N_PROCS = 2

    def test_multidim(self):
        # this test would error out before the fix
        class Adder(om.ExplicitComponent):
            def initialize(self):
                self.options.declare('n0')

            def setup(self):
                n0 = self.options['n0']
                self.add_input('x', shape_by_conn=True, distributed=True)
                self.add_output('x_sum', shape=1)

            def compute(self, inputs, outputs):
                outputs['x_sum'] = self.comm.allreduce(np.sum(inputs['x']))


        n0 = (20,2) if MPI.COMM_WORLD.rank ==0 else (8,2)
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc',om.IndepVarComp())
        ivc.add_output('x', val = np.random.random(n0), distributed=True)

        p.model.add_subsystem('adder', Adder(n0=n0))
        p.model.connect('ivc.x','adder.x')

        p.setup()
        p.run_model()

        inputs_report(p, outfile='temp_inputs_report.md', display=True, precision=6, title=None, tablefmt='github')



"""
Generate report on the results of the optimization.
"""
import base64
import datetime
from functools import partial
import io
import os
import pathlib
import sys
import time

import numpy as np

from openmdao.utils.reports_system import register_report

try:
    import matplotlib as mpl
    from matplotlib import pyplot as plt
    from matplotlib import patches
except ImportError:
    mpl = None


from openmdao.core.constants import INF_BOUND
from openmdao.utils.mpi import MPI
from openmdao.utils.om_warnings import issue_warning, DriverWarning
from openmdao.visualization.tables.table_builder import generate_table


# Report file constants
_default_optimizer_report_filename = 'opt_report.html'
_optimizer_report_template = 'opt_report_template.html'

MAX_VARNAME_LENGTH = 48

# value tolerances
_bounds_tolerance = 1.0E-4
_near_bounds_tolerance = 1.0E-4
_equality_constraint_tolerance = 1e-4

# Colors
#     from https://jfly.uni-koeln.de/color/ some colorblind friendly colors
_out_of_bounds_test_color = 'rgb(213, 94, 0)'
_out_of_bounds_plot_color = (0.83529, 0.36862, 0.)
_out_of_bounds_plot_alpha = 0.2

_out_of_bounds_plot_hatch_pattern = 'xxxxx'
_out_of_bounds_constraint_visual_hatch_pattern = 'xxxxxxxxx'

_out_of_bounds_hatch_color = (0., 0., 0.)
_out_of_bounds_hatch_width = 0.3

_in_bounds_plot_color = (0., 0.61960, 0.45098)
_in_bounds_plot_alpha = 0.2
_in_bounds_constraint_visual_alpha = 0.2

_value_plot_color = 'black'
_plot_marker_size = 0.1
_plot_value_linewidth = 0.5
_equality_constraint_dot_size = 3

# overall image parameters and layout
_sparkline_figsize = (3, .5)
_scalar_visual_figsize = (2.0, .2)
_plot_dpi = 150
_plot_pad_inches = 0
_plot_padding_fraction = 0.2

# Font sizes
_sparkline_font_size = 7
_scalar_visual_font_size = 4

# variable value formatting
_variable_label_format = '5.3g'

# https://jfly.uni-koeln.de/color/
# scalar visual parameters
_plot_x_max = 1.0
_plot_y_max = 1.0
_plot_y_margin = 0.2
_lower_plot = _plot_x_max / 3.
_upper_plot = 2 * _plot_x_max / 3.
_pointer_half_width = 0.04
_pointer_height = 0.4
_text_height = 0.3
_pointer_line_width = 0.05
_near_bound_highlight_half_width = 0.05
_near_bound_highlight_half_width_y_min = 0.0
_near_bound_highlight_half_width_y_max = _plot_y_max
_near_bound_highlight_alpha = 0.7
_equality_bound_width = 0.01
# For the ellipsis that are displayed when scalar is way outside of bounds
_ellipsis_x_offset = 0.2
_ellipse_width = 0.01
_ellipse_height = 0.15


def _optimizer_report_register():
    register_report('optimizer', opt_report, 'Summary of optimization',
                    'Problem', 'run_driver', 'post')


def opt_report(prob, outfile=None):
    """
    Write a summary of the optimization to the given file.

    Parameters
    ----------
    prob : Problem
        The Problem for which the optimization report is being generated.
    outfile : str or None
        The path to the HTML file to be written.  If None (default), write to the default report
        output path.
    """
    driver = prob.driver
    if not driver.supports['optimization']:
        driver_class = type(driver).__name__
        issue_warning(f"The optimizer report is not applicable for Driver type '{driver_class}', "
                      "which does not support optimization", category=DriverWarning)
        return

    # only create report on rank 0
    create = MPI is None or MPI.COMM_WORLD.rank == 0

    if not outfile:
        outfile = _default_optimizer_report_filename

    outfilepath = pathlib.Path(prob.get_reports_dir(force=True)).joinpath(outfile)

    driver_scaling = True

    # Collect data from the problem
    abs2prom = prob.model._var_abs2prom

    def get_prom_name(abs_name):
        if abs_name in abs2prom['input']:
            return abs2prom['input'][abs_name]
        elif abs_name in abs2prom['output']:
            return abs2prom['output'][abs_name]
        else:
            return abs_name

    # Collect the entire array of array valued desvars and constraints (ignore indices)
    objs_vals = {}
    desvars_vals = {}
    cons_vals = {}

    objs_meta = {}
    desvars_meta = {}
    cons_meta = {}

    with prob.model._scaled_context_all():
        for abs_name, meta in driver._objs.items():
            prom_name = get_prom_name(abs_name)
            objs_meta[prom_name] = meta
            objs_vals[prom_name] = \
                driver.get_objective_values(driver_scaling=driver_scaling)[abs_name]

        for abs_name, meta in driver._designvars.items():
            prom_name = get_prom_name(abs_name)
            desvars_meta[prom_name] = meta
            desvars_vals[prom_name] = \
                driver.get_design_var_values(driver_scaling=driver_scaling)[abs_name]

        for abs_name, meta in prob.driver._cons.items():
            if meta.get('alias') is not None:
                prom_name = abs_name
            else:
                prom_name = get_prom_name(abs_name)
            cons_meta[prom_name] = meta
            cons_vals[prom_name] = \
                driver.get_constraint_values(driver_scaling=driver_scaling)[abs_name]

    header_html = _make_header_table(prob)

    objs_html = _make_obj_table(objs_meta=objs_meta, objs_vals=objs_vals,
                                cols=['val', 'ref', 'ref0', 'adder', 'scaler', 'units'])

    desvars_html = _make_dvcons_table(meta_dict=desvars_meta, vals_dict=desvars_vals, kind='desvar',
                                      cols=['min', 'max', 'mean', 'lower', 'upper', 'equals', 'ref',
                                            'ref0', 'units', 'visual'])

    cons_html = _make_dvcons_table(meta_dict=cons_meta, vals_dict=cons_vals, kind='constraint',
                                   cols=['min', 'max', 'mean', 'lower', 'upper', 'equals', 'ref',
                                         'ref0', 'units', 'visual'])

    driver_info_html = _make_opt_value_table(driver)

    this_dir = os.path.dirname(os.path.abspath(__file__))

    if create:
        with open(os.path.join(this_dir, _optimizer_report_template), 'r', encoding='utf-8') as f:
            template = f.read()

        with open(outfilepath, 'w') as f:
            s = template.format(title=prob._name, header=header_html, objs=objs_html,
                                desvars=desvars_html, cons=cons_html, driver=driver_info_html,
                                driver_class=type(driver).__name__)
            f.write(s)


def _make_header_table(prob):
    """
    Make the HTML table at the top of the report with basic info about the optimization run.

    Parameters
    ----------
    prob : OpenMDAO Problem object
        The report will be run on this Problem.

    Returns
    -------
    str
        HTML table that displays basic optimization run info.
    """
    t = datetime.datetime.now()
    time_stamp = t.strftime("%Y-%m-%d %H:%M:%S %Z")
    runtime = prob.driver.opt_result['runtime']
    runtime_ms = (runtime * 1000.0) % 1000.0
    runtime_formatted = \
        f"{time.strftime('%H hours %M minutes %S seconds', time.gmtime(runtime))} " \
        f"{runtime_ms:.1f} milliseconds"

    rows = list()
    rows.append(['Problem:', prob._name])
    rows.append(['Script:', sys.argv[0]])
    rows.append(['Optimizer:', prob.driver._get_name()])
    rows.append(['Number of driver iterations:', prob.driver.opt_result['iter_count']])
    rows.append(['Number of objective calls:', prob.driver.opt_result['obj_calls']])
    rows.append(['Number of derivative calls:', prob.driver.opt_result['deriv_calls']])
    rows.append(['Execution start time:', time_stamp])
    rows.append(['Wall clock run time:', runtime_formatted])
    rows.append(['Exit status:', prob.driver.opt_result['exit_status']])

    return generate_table(rows, tablefmt='html')


def _make_opt_value_table(driver):
    """
    Make the HTML table that summarizes the optimizer settings.

    Parameters
    ----------
    driver : OpenMDAO Driver object
        The table will include info on the settings for this driver.

    Returns
    -------
    str
        HTML table that displays driver settings info.
    """
    opt_settings = []
    for key, meta in driver.options.items():
        meta = driver.options._dict[key]
        opt_settings.append((key, meta['val'], meta['desc']))
    opt_settings_table = generate_table(opt_settings, headers=['Setting', 'Val', 'Description'],
                                        tablefmt='html')

    html = ''
    if opt_settings_table is not None:
        html += f'\n{opt_settings_table}'

    return html


def _make_obj_table(objs_meta, objs_vals,
                    cols=['size', 'index', 'val', 'ref', 'ref0', 'adder', 'scaler', 'units']):
    """
    Make a table of info about the objective.

    Parameters
    ----------
    objs_meta : dict
        Dictionary of metadata about the objectives.
    objs_vals : dict
        Dictionary of values for the objectives.
    cols : list
        List of columns to be displayed in the table.

    Returns
    -------
    str
        An HTML image tag containing the table holding the info about the variable.
    """
    _col_names = cols if cols is not None else []
    col_names = ['name'] + _col_names

    # Get the values for all the elements in the tables
    rows = []
    for prom_name, meta in objs_meta.items():
        row = {}
        if meta['ref0'] is None and meta['ref'] is None and meta['scaler'] is None and \
                meta['adder'] is None:
            meta['ref'] = 1.0
            meta['ref0'] = 0.0
        for col_name in col_names:
            if col_name == 'name':
                row[col_name] = prom_name
                if len(row[col_name]) > MAX_VARNAME_LENGTH:
                    split_name = row[col_name].split('.')
                    row[col_name] = '.'.join(split_name[:2]) + '...' + split_name[-1]
            elif col_name == 'val':
                row[col_name] = objs_vals[prom_name]
            elif col_name == 'units':
                if meta['units'] is not None:
                    row[col_name] = meta[col_name]
                else:
                    row[col_name] = objs_meta[prom_name][col_name]
            else:
                row[col_name] = meta[col_name]
        rows.append(row)

    return generate_table(rows, headers='keys', tablefmt='html')


def _make_dvcons_table(meta_dict, vals_dict, kind,
                       cols=['lower', 'upper', 'ref', 'ref0', 'adder', 'scaler', 'units', 'min',
                             'max', 'visual']):
    """
    Make a table of info about either design variables or constraints.

    Parameters
    ----------
    meta_dict : dict
        Dictionary of metadata about the variables.
    vals_dict : dict
        Dictionary of values for the variables.
    kind : str, must be 'desvar' or 'constraint'
        Indicates whether table is for 'desvar' or 'constraint'.
    cols : list
        List of columns to be displayed in the table.

    Returns
    -------
    str
        An HTML image tag containing the table holding the info about the variable.
    """
    _col_names = cols if cols is not None else []
    col_names = ['name', 'alias', 'size'] + _col_names

    # Get the values for all the elements in the tables
    rows = []

    for name, meta in meta_dict.items():

        row = {}

        if meta['ref0'] is None and meta['ref'] is None and meta['scaler'] is None and \
                meta['adder'] is None:
            meta['ref'] = 1.0
            meta['ref0'] = 0.0

        alias = meta.get('alias', '')

        # the scipy optimizer, when using COBYLA, creates constraints under the hood.
        #  But the values are not given by the driver, so use this as a sign that this
        #  variable should be skipped
        if name not in vals_dict:
            continue

        for col_name in col_names:
            if col_name == 'name':
                row[col_name] = name
                if len(row[col_name]) > MAX_VARNAME_LENGTH:
                    split_name = row[col_name].split('.')
                    row[col_name] = '.'.join(split_name[:2]) + '...' + split_name[-1]
            elif col_name == 'alias':
                row[col_name] = alias
            elif col_name == 'mean':
                mean_val = np.mean(vals_dict[name])
                row[col_name] = _indicate_value_is_derived_from_array(mean_val, vals_dict[name])
            elif col_name == 'min':
                if isinstance(vals_dict[name], np.ndarray):
                    min_val = min(vals_dict[name])  # get min. Could be an array
                    min_val_as_str = _indicate_value_is_derived_from_array(min_val, vals_dict[name])
                else:
                    min_val_as_str = str(vals_dict[name])
                comp = (vals_dict[name] - meta['lower']) < _bounds_tolerance
                if np.any(comp):
                    row[col_name] = \
                        f'<span style="color:{_out_of_bounds_test_color}">({min_val_as_str})</span>'
                else:
                    row[col_name] = min_val_as_str
            elif col_name == 'max':
                if isinstance(vals_dict[name], np.ndarray):
                    max_val = max(vals_dict[name])  # get max. Could be an array
                    max_val_as_str = _indicate_value_is_derived_from_array(max_val, vals_dict[name])
                else:
                    max_val_as_str = str(vals_dict[name])
                comp = (meta['upper'] - vals_dict[name]) < _bounds_tolerance
                if np.any(comp):
                    row[col_name] = \
                        f'<span style="color:{_out_of_bounds_test_color}">({max_val_as_str})</span>'
                else:
                    row[col_name] = max_val_as_str
            elif col_name == 'lower':
                if np.all(meta[col_name] == -INF_BOUND):
                    row[col_name] = None
                else:
                    lower_val = np.mean(meta[col_name])
                    row[col_name] = _indicate_value_is_derived_from_array(lower_val, meta[col_name])
            elif col_name == 'upper':
                if np.all(meta[col_name] == INF_BOUND):
                    row[col_name] = None
                else:
                    upper_val = np.mean(meta[col_name])
                    row[col_name] = _indicate_value_is_derived_from_array(upper_val, meta[col_name])
            elif col_name == 'equals':
                if 'equals' not in meta or meta['equals'] is None:
                    row[col_name] = ''
                else:
                    equals_val = np.mean(meta[col_name])
                    row[col_name] = _indicate_value_is_derived_from_array(equals_val,
                                                                          meta[col_name])
            elif col_name == 'units':
                if meta['units'] is not None:
                    row[col_name] = meta[col_name]
                else:
                    if alias:
                        row[col_name] = meta_dict[alias][col_name]
                    else:
                        row[col_name] = meta_dict[name][col_name]
            elif col_name == 'visual':
                if mpl:
                    val = vals_dict[name]
                    if np.isscalar(val) or val.shape == (1,):
                        row['visual'] = _constraint_plot(meta=meta, val=vals_dict[name], kind=kind)
                    else:
                        row['visual'] = _sparkline(meta=meta, val=vals_dict[name], kind=kind)
                else:
                    row['visual'] = \
                        '<span class="plot-unavailable">Visuals require matplotlib</span>'
            elif col_name == 'size':
                row[col_name] = int(meta[col_name])  # sometimes size in the meta data is a numpy
                # array so generate_table does different formatting for that
            elif col_name in ['ref', 'ref0']:
                if meta[col_name] is not None:
                    derived_val = np.mean(meta[col_name])
                    row[col_name] = _indicate_value_is_derived_from_array(derived_val,
                                                                          meta[col_name])
                else:
                    row[col_name] = 'None'
            else:
                row[col_name] = meta[col_name]

        rows.append(row)

    return generate_table(rows, headers='keys', tablefmt='html', safe=False, precision='.4e')


def _sparkline(kind, meta, val, width=300):
    """
    Given the metadata & value of design variable or constraint, make an html-embeddable sparkline.

    Parameters
    ----------
    kind : str
        One of 'desvar' or 'constraint' to specify which type of sparkline is being made.
        This has a slight impact on how the bounds are plotted.
    meta : dict-like
        The metadata associated with the design variable or constraint.
    val : np.array
        The value of the design variable or constraint.
    width : int
        The width of the figure in the returned HTML tag.

    Returns
    -------
    str
        An HTML image tag containing the encoded sparkline image.
    """
    # Prepare the matplotlib figure/axes
    _backend = mpl.get_backend()  # Save it so we can set it back at the end
    mpl.use('Agg')

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=_sparkline_figsize,
                           tight_layout=False, dpi=_plot_dpi)

    # settings across all plots
    ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(f"%{_variable_label_format}"))
    ax.tick_params(axis='x', labelsize=_sparkline_font_size)
    ax.tick_params(axis='y', labelsize=_sparkline_font_size)

    try:
        if kind == 'desvar':
            _desvar_or_ineq_constraint_sparkline(ax, meta, val)
        elif 'equals' not in meta or meta['equals'] is None:
            _desvar_or_ineq_constraint_sparkline(ax, meta, val)
        else:
            _eq_constraint_sparkline(ax, meta, val)
    except (ValueError, IndexError):
        mpl.use(_backend)  # set it back
        plt.close()
        return '<span class="plot-unavailable">Plot unavailable</span>'

    tmpfile = io.BytesIO()
    fig.patch.set_alpha(0.0)  # So that the figures are transparent
    fig.savefig(tmpfile, format='png', bbox_inches='tight', pad_inches=_plot_pad_inches)
    encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')
    html = f'<img width={width} src=\'data:image/png;base64,{encoded}\'>'

    mpl.use(_backend)  # set it back
    plt.close()

    return html


def _desvar_or_ineq_constraint_sparkline(ax, meta, val):
    """
    Use matplotlib to plot a visual showing the values of the desvar or constraint and any bounds.

    Parameters
    ----------
    ax : Matplotlib Axes instance
        Contains most of the figure elements.
    meta : dict
        The metadata associated with the design variable.
    val : np.array
        The value of the design variable.
    """
    indices = np.s_[...] if meta['indices'] is None else meta['indices']()

    _val = np.asarray(val).ravel()

    # get array for the lower and upper bounds
    if isinstance(meta['lower'], float):
        _lower = (meta['lower'] * np.ones_like(val)[indices]).ravel()
    else:
        _lower = np.asarray(meta['lower']).ravel()

    if isinstance(meta['upper'], float):
        _upper = (meta['upper'] * np.ones_like(val)[indices]).ravel()
    else:
        _upper = np.asarray(meta['upper']).ravel()

    # Get the x values and their range
    ar = np.arange(_val.size)
    x_min = 0
    x_max = val.size - 1

    _lower, _lower_no_inf_min, _lower_no_inf_max = _get_bound_array_min_max(_lower, indices)
    _upper, _upper_no_inf_min, _upper_no_inf_max = _get_bound_array_min_max(_upper, indices)

    # get y min and max
    y_min = min(_lower_no_inf_min, _upper_no_inf_min, np.min(_val))
    y_max = max(_lower_no_inf_max, _upper_no_inf_max, np.max(_val))

    # add a little to the top and bottom of the plot
    yrange = y_max - y_min
    ymin_plot = y_min - yrange * _plot_padding_fraction
    ymax_plot = y_max + yrange * _plot_padding_fraction

    xrange = x_max - x_min
    xmin_plot = x_min - xrange * _plot_padding_fraction
    xmax_plot = x_max + xrange * _plot_padding_fraction

    # set plot limits
    ax.set_xlim([xmin_plot, xmax_plot])
    ax.set_ylim([ymin_plot, ymax_plot])

    # set ticks
    ax.set_xticks([x_min, x_max])
    ax.set_yticks([y_min, y_max])

    # plot actual values
    ax.plot(ar, _val, '-o', markersize=_plot_marker_size, color=_value_plot_color,
            linewidth=_plot_value_linewidth)

    # Need to do this because of a bug in matplotlib. If the upper or lowers include INF_BOUND
    #  it affects how the other side of the fill_between is drawn
    _lower = np.clip(_lower, ymin_plot, ymax_plot)
    _upper = np.clip(_upper, ymin_plot, ymax_plot)

    # plot lower area, if exists
    if not (isinstance(meta['lower'], float) and meta['lower'] == -INF_BOUND):
        ax.fill_between(ar, ymin_plot, _lower, color=_out_of_bounds_plot_color,
                        hatch=_out_of_bounds_plot_hatch_pattern, alpha=_out_of_bounds_plot_alpha)
        ax.plot(ar, _lower, color=_out_of_bounds_plot_color, linewidth=_plot_value_linewidth)

    # plot upper area, if exists
    if not (isinstance(meta['upper'], float) and meta['upper'] == INF_BOUND):
        ax.fill_between(ar, _upper, ymax_plot, color=_out_of_bounds_plot_color,
                        hatch=_out_of_bounds_plot_hatch_pattern, alpha=_out_of_bounds_plot_alpha)
        ax.plot(ar, _upper, color=_out_of_bounds_plot_color, linewidth=_plot_value_linewidth)

    # Plot area where bounds are satisfied
    ax.fill_between(ar, _lower, _upper, color=_in_bounds_plot_color, alpha=_in_bounds_plot_alpha)


def _eq_constraint_sparkline(ax, meta, val):
    """
    Plot to matplotlib a visual showing the values of the equality constraint and also the value.

    Parameters
    ----------
    ax : Matplotlib Axes instance
        Contains most of the figure elements.
    meta : dict
        The metadata associated with the variables, including info about the constraint.
    val : np.array
        The value of the variable.
    """
    indices = np.s_[...] if meta['indices'] is None else meta['indices']

    # get value array flattened
    _val = np.asarray(val).ravel()

    # get equal constraint array
    if 'equals' not in meta:
        raise ValueError("Equality constraint sparkline cannot be "
                         "drawn without equality constraint value")
    elif isinstance(meta['equals'], float):
        _equals = (meta['equals'] * np.ones_like(val)[indices]).ravel()
    else:
        _equals = np.asarray(meta['equals']).ravel()

    # get x coordinate array and info
    ar = np.arange(val.size)
    x_min = 0
    x_max = val.size - 1
    xrange = x_max - x_min
    xmin_plot = x_min - xrange * _plot_padding_fraction
    xmax_plot = x_max + xrange * _plot_padding_fraction
    ax.set_xlim([xmin_plot, xmax_plot])
    ax.set_xticks([x_min, x_max])

    # Get y coordinate plotting info
    y_min = min(np.min(_val), np.min(_equals))
    y_max = max(np.max(_val), np.max(_equals))
    yrange = y_max - y_min
    ymin_plot = y_min - yrange * _plot_padding_fraction
    ymax_plot = y_max + yrange * _plot_padding_fraction
    ax.set_ylim([ymin_plot, ymax_plot])
    ax.set_yticks([y_min, y_max])

    # plot the constraint as a line
    ax.plot(ar, _equals, color=_value_plot_color, linewidth=_plot_value_linewidth)

    # plot the actual values as dots colored by whether they satisfy the constraint
    err = _val - _equals
    colors = []
    for e in err:
        if np.abs(e) < _equality_constraint_tolerance:
            colors.append(_in_bounds_plot_color)
        else:
            colors.append(_out_of_bounds_plot_color)
    ax.scatter(ar, _val, color=colors, s=_equality_constraint_dot_size)


def _constraint_plot(kind, meta, val, width=300):
    """
    Given the metadata and value of a design variable or constraint, make an html-embeddable plot.

    Only for scalars

    Parameters
    ----------
    kind : str
        One of 'desvar' or 'constraint' to specify which type of sparkline is being made.
        This has a slight impact on how the bounds are plotted.
    meta : dict-like
        The metadata associated with the design variable or constraint.
    val : np.array
        The value of the design variable or constraint.
    width : int
        The width of the figure in the returned HTML tag.

    Returns
    -------
    str
        An HTML image tag containing the encoded sparkline image.
    """
    if not (np.isscalar(val) or val.shape == (1,)):
        raise ValueError("Value for the _constraint_plot function must be a "
                         f"scalar. Variable {meta['name']} is not a scalar")

    if kind == 'desvar' and meta['upper'] == INF_BOUND and meta['lower'] == -INF_BOUND:
        return   # nothing to plot

    # Equality constraints are visualized differently
    equals = meta['equals'] if 'equals' in meta else None
    if equals is not None:
        if abs(val - equals) < _equality_constraint_tolerance:
            html = '<span class="equality-constraint equality-constraint-satisfied">&#10003;</span>'
        else:
            html = '<span class="equality-constraint equality-constraint-violated">&#10007;</span>'
        return html

    # If lower and upper are the same value, visualize the same as an equality constraint
    if (
        'lower' in meta and meta['lower'] != -INF_BOUND
        and
        'upper' in meta and meta['upper'] != INF_BOUND
        and
        meta['lower'] == meta['upper']
    ):
        if abs(val - meta['lower']) < _equality_constraint_tolerance:
            html = '<span class="equality-constraint equality-constraint-satisfied">&#10003;</span>'
        else:
            html = '<span class="equality-constraint equality-constraint-violated">&#10007;</span>'
        return html

    _backend = mpl.get_backend()
    mpl.use('Agg')

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=_scalar_visual_figsize, dpi=_plot_dpi)

    var_bounds_plot(kind, ax, float(val), meta['lower'], meta['upper'])
    tmpfile = io.BytesIO()
    fig.savefig(tmpfile, format='png', transparent=True, bbox_inches='tight',
                pad_inches=_plot_pad_inches)
    encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')

    html = f'<img width={width} src=\'data:image/png;base64,{encoded}\'>'

    mpl.use(_backend)
    plt.close()

    return html


def _indicate_value_is_derived_from_array(derived_value, original_value):
    """
    Given a value or bound, & its derived value for use in report, format the output for the tables.

    Parameters
    ----------
    derived_value : float
        The value of the variable to be shown in the table.
    original_value : float or np.ndarray
        The original value of the variable. Need this to know if it was a scalar.

    Returns
    -------
    str
        A formatted string to be used to display the value of this variable in the tables.
    """
    displayed_string = f'{derived_value:{_variable_label_format}}'.strip()
    if isinstance(original_value, np.ndarray) and original_value.size > 1:
        out = '|{}|'.format(displayed_string)
    else:
        out = displayed_string
    return out


def _get_bound_array_min_max(bounds, indices):
    """
    Given a bounds (either lower or upper) & indices, return the full array bounds, min,& max.

    Parameters
    ----------
    bounds : float
        An array representing either lower or upper bounds of a desvar or constraint.
    indices : sequence of int
        If variable is an array, these indicate which entries are of
        interest for this particular response.

    Returns
    -------
    _bounds
        An array representing the bounds.
    _bounds_min
        The minimum of the bounds, excluding any entries in the _bounds array that
          equal -INF_BOUND or INF_BOUND. If all the values are those, return INF_BOUND
          so that when mins are taken that include this value, it doesn't affect the result
    _bounds_max
        The maximum of the bounds, excluding any entries in the _bounds array that
          equal -INF_BOUND or INF_BOUND. If all the values are those, return -INF_BOUND
          so that when maxes are taken that include this value, it doesn't affect the result
    """
    _bounds_no_inf = bounds[np.where(((bounds != -INF_BOUND) & (bounds != INF_BOUND)))]
    if _bounds_no_inf.size > 0:
        _bounds_min = np.min(_bounds_no_inf)
        _bounds_max = np.max(_bounds_no_inf)
    else:  # so that when we do min and max on these they are not involved in getting the min/max
        _bounds_min = INF_BOUND
        _bounds_max = - INF_BOUND

    if isinstance(bounds, float):
        _bounds = (bounds * np.ones_like(bounds)[indices]).ravel()
    else:
        _bounds = np.asarray(bounds).ravel()

    return _bounds, _bounds_min, _bounds_max


def var_bounds_plot(kind, ax, value, lower, upper):
    """
    Make a plot to show where a variable is relative to constraints.

    Parameters
    ----------
    kind : str
        One of 'desvar' or 'constraint' to specify which type of plot is being made.
    ax : Matplotlib Axes instance
        Contains most of the figure elements.
    value : float
        The design var value.
    lower : float or None
        Lower constraint.
    upper : float or None
        Upper constraint.
    """
    # must handle 5 cases if both upper and lower are given:
    #  - value much less than lower
    #  - value a little less than lower
    #  - value between lower and upper
    #  - value a little greater than upper
    #  - value much greater than upper

    # also need to handle one-sided constraints where only one of lower and upper is given
    if kind == 'constraint' and upper == INF_BOUND and lower == -INF_BOUND:
        raise ValueError("Upper and lower bounds cannot all be None for a constraint")

    # Basic plot setup
    plt.rcParams['hatch.linewidth'] = _out_of_bounds_hatch_width  # can't seem to do this any other
    #                                                              way. Cannot control hatch
    #                                                              pattern in Rectangle with
    #                                                              set_linewidth
    plt.rcParams['hatch.color'] = _out_of_bounds_plot_color
    ax.set_axis_off()
    ax.set_xlim([-_pointer_half_width, _plot_x_max + _pointer_half_width])
    ax.set_ylim(-_pointer_height - _text_height - _plot_y_margin,
                _plot_y_max + _text_height + _plot_y_margin)

    func_val_to_plot_coord = partial(_val_to_plot_coord, lower=lower, upper=upper)
    value_in_plot_coord = func_val_to_plot_coord(value)

    if upper == INF_BOUND:  # there is a lower bound
        _draw_in_or_out_bound_section(ax, 0, _plot_x_max / 2, False)
        _draw_in_or_out_bound_section(ax, _plot_x_max / 2, _plot_x_max / 2, True)
        _draw_boundary_label(ax, _plot_x_max / 2,
                             "lower = " + f"{lower:{_variable_label_format}}".strip())

        if abs(value - lower) < _near_bounds_tolerance:
            pointer_plot_coord = _plot_x_max / 2
            _draw_bound_highlight(ax, _plot_x_max / 2)
            pointer_color = _in_bounds_plot_color
        elif value >= lower:
            pointer_plot_coord = 3. * _plot_x_max / 4
            pointer_color = _in_bounds_plot_color
        else:
            pointer_color = _out_of_bounds_plot_color
            pointer_plot_coord = 1. * _plot_x_max / 4
        _draw_pointer_and_label(ax, pointer_plot_coord, pointer_color, value)
        return

    if lower == -INF_BOUND:  # there is an upper bound
        _draw_in_or_out_bound_section(ax, 0, _plot_x_max / 2, True)
        _draw_in_or_out_bound_section(ax, _plot_x_max / 2, _plot_x_max / 2, False)
        _draw_boundary_label(ax, _plot_x_max / 2,
                             "upper = " + f"{upper:{_variable_label_format}}".strip())

        if abs(value - upper) < _near_bounds_tolerance:  # value is close to bound
            pointer_plot_coord = _plot_x_max / 2
            _draw_bound_highlight(ax, _plot_x_max / 2)
            pointer_color = _in_bounds_plot_color
        elif value <= upper:  # value satisfies bound
            pointer_plot_coord = 1. * _plot_x_max / 4
            pointer_color = _in_bounds_plot_color
        else:  # value violates the bound
            pointer_plot_coord = 3. * _plot_x_max / 4
            pointer_color = _out_of_bounds_plot_color

        _draw_pointer_and_label(ax, pointer_plot_coord, pointer_color, value)
        return

    # If we get this far, there are both lower and upper bounds

    # in bounds visual is always the same
    _draw_in_or_out_bound_section(ax, _lower_plot, _upper_plot - _lower_plot, True)

    # draw below bound
    if value_in_plot_coord >= 0.0:
        _draw_in_or_out_bound_section(ax, 0, _lower_plot, False)
    else:  # value is off to the left of the plot
        _draw_in_or_out_bound_section(ax, _lower_plot / 3., 2 * _lower_plot / 3., False)
        _draw_ellipsis(ax, 0.0)

    # draw upper bound
    if value_in_plot_coord <= _plot_x_max:
        _draw_in_or_out_bound_section(ax, _upper_plot, _lower_plot, False)
    else:  # value is off to the right of the plot
        _draw_in_or_out_bound_section(ax, _upper_plot, 2 * _lower_plot / 3., False)
        _draw_ellipsis(ax, _upper_plot + 2 * _lower_plot / 3)

    # draw upper and lower labels
    _draw_boundary_label(ax, func_val_to_plot_coord(lower), str(lower))
    _draw_boundary_label(ax, func_val_to_plot_coord(upper), str(upper))

    # add highlight if value near a bound
    if abs(value - lower) / abs(upper - lower) < _near_bounds_tolerance:
        _draw_bound_highlight(ax, func_val_to_plot_coord(lower))
    elif abs(value - upper) / abs(upper - lower) < _near_bounds_tolerance:
        _draw_bound_highlight(ax, func_val_to_plot_coord(upper))

    # pointer and pointer label
    if value_in_plot_coord < 0.0:
        pointer_plot_coord = 0.0
    elif value_in_plot_coord > _plot_x_max:
        pointer_plot_coord = _plot_x_max
    else:
        pointer_plot_coord = value_in_plot_coord
    pointer_color = _in_bounds_plot_color if (lower <= value <= upper) \
        else _out_of_bounds_plot_color
    _draw_pointer_and_label(ax, pointer_plot_coord, pointer_color, value)


# A series of functions used to draw parts of the scalar constraints visual
def _draw_in_or_out_bound_section(ax, x_left, width, is_in_bound):
    if is_in_bound:
        color = _in_bounds_plot_color
        hatch = None
        alpha = _in_bounds_plot_alpha
    else:
        color = _out_of_bounds_plot_color
        hatch = _out_of_bounds_constraint_visual_hatch_pattern
        # hatch = None
        alpha = _out_of_bounds_plot_alpha
    rectangle = patches.Rectangle((x_left, 0), width, _plot_y_max, facecolor=color,
                                  hatch=hatch, alpha=alpha)
    ax.add_patch(rectangle)


def _draw_bound_highlight(ax, x):
    rectangle = patches.Rectangle((
        x - _near_bound_highlight_half_width, _near_bound_highlight_half_width_y_min),
        2 * _near_bound_highlight_half_width, _near_bound_highlight_half_width_y_max,
        edgecolor='black', facecolor='yellow', alpha=_near_bound_highlight_alpha)
    ax.add_patch(rectangle)


def _draw_ellipsis(ax, x_left):
    # Draw three dots as an ellipsis to show that the value is beyond
    #   either the left or right edge of the plot
    for i in [1, 2, 3]:
        circle = patches.Ellipse((x_left + i * _lower_plot / 12., _ellipsis_x_offset),
                                 _ellipse_width, _ellipse_height,
                                 facecolor=_out_of_bounds_plot_color)
        ax.add_patch(circle)


def _draw_boundary_label(ax, pointer_plot_coord, s):
    ax.text(pointer_plot_coord, _plot_y_max + _text_height,
            s,
            horizontalalignment='center',
            verticalalignment='bottom',
            size=_scalar_visual_font_size
            )


def _draw_pointer_and_label(ax, pointer_plot_coord, pointer_color, value):
    pts = np.array([
        [pointer_plot_coord - _pointer_half_width, -_pointer_height],
        [pointer_plot_coord + _pointer_half_width, -_pointer_height],
        [pointer_plot_coord, 0.0]
    ])
    p = patches.Polygon(pts, closed=True, facecolor=pointer_color, edgecolor='black',
                        linewidth=_pointer_line_width)
    ax.add_patch(p)
    plt.text(pointer_plot_coord, -_pointer_height - _text_height,
             f"{value:{_variable_label_format}}".strip(),
             horizontalalignment='center', verticalalignment='top', size=_scalar_visual_font_size)


def _val_to_plot_coord(value, lower, upper):
    # need to get function that maps actual values to 0.0 to 1.0
    # and where lower maps to 1./3 and upper to 2/3
    # Used with Python's functools to make a Python function out of this
    plot_coord = 1. / 3. + (value - lower) / (upper - lower) * 1. / 3.
    if isinstance(plot_coord, np.ndarray):
        return plot_coord[0]
    return plot_coord



"""
This test assures that the use of nonstandard datatypes anywhere that they are allowed in a model
does not break the recording of the JSON data structures needed for the model viewer.
"""

import unittest

import numpy as np

import openmdao.api as om
from openmdao.core.driver import Driver
from openmdao.utils.testing_utils import use_tempdirs


class BadOpt(object):
    def junk(self):
        pass


class NonSerComp(om.ExplicitComponent):

    def setup(self):
        self.add_input('x', np.zeros((6, )),
                       tags=['a', 'c'])
        self.add_output('y', np.zeros((6, )),
                       tags=['a', 'b'])

        self.add_discrete_input('dx', [{('Discrete_i', BadOpt): (2, {((1, ), (2, )): 'stuff'})}])
        self.add_discrete_output('dy', [{('Discrete_o', BadOpt): (2, {((1, ), (2, )): 'stuff'})}])
        self.add_discrete_input('dcomplex', 3 + 5j)

    def initialize(self):
        self.options.declare('good', 'good_string')
        self.options.declare('bad', [{(1, BadOpt): (2, 3)}])
        self.options.declare('bad2', {((1, ), (2, )): 'stuff'})
        self.options.declare('nonrec', 3.0, recordable=False)
        self.options.declare('cx', 3 + 7j)


class NonSerIComp(om.ImplicitComponent):

    def setup(self):
        self.add_input('xx', np.zeros((6, )))
        self.add_output('yy', np.zeros((6, )))

        self.add_discrete_input('problem', None)

    def initialize(self):
        self.options.declare('good', 'good_string')
        self.options.declare('bad', [{(1, BadOpt): (2, 3)}])
        self.options.declare('bad2', {((1, ), (2, )): 'stuff'})
        self.options.declare('nonrec', 3.0, recordable=False)
        self.options.declare('problem')


class NonSerNL(om.NonlinearRunOnce):

    def _declare_options(self):
        super()._declare_options()
        self.options.declare('bad', [{(1, BadOpt): (2, 3)}])
        self.options.declare('bad2', {((1, ), (2, )): 'stuff'})
        self.options.declare('nonrec', 3.0, recordable=False)


class NonSerLN(om.LinearRunOnce):

    def _declare_options(self):
        super()._declare_options()
        self.options.declare('bad', [{(1, BadOpt): (2, 3)}])
        self.options.declare('bad2', {((1, ), (2, )): 'stuff'})
        self.options.declare('nonrec', 3.0, recordable=False)


class NonSerDriver(Driver):

    def _declare_options(self):
        super()._declare_options()
        self.options.declare('bad', [{(1, BadOpt): (2, 3)}])
        self.options.declare('bad2', {((1, ), (2, )): 'stuff'})
        self.options.declare('nonrec', 3.0, recordable=False)


@use_tempdirs
class TestSerialization(unittest.TestCase):

    def test_exhaustive_model(self):
        em_prob = om.Problem()
        em_model = em_prob.model
        em_model.add_subsystem('ns', NonSerComp())
        em_model.add_subsystem('nsi', NonSerIComp())
        em_prob.setup()
        em_prob.final_setup()

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('ns', NonSerComp())
        nsi = model.add_subsystem('nsi', NonSerIComp())

        nsi.options['problem'] = em_prob

        model.nonlinear_solver = NonSerNL()
        model.linear_solver = NonSerLN()
        nsi.nonlinear_solver = NonSerNL()
        nsi.linear_solver = NonSerLN()

        model.add_design_var('ns.x', indices=om.slicer[2:])
        model.add_design_var('ns.dx')
        model.add_constraint('ns.y', indices=om.slicer[2:])
        model.add_objective('nsi.yy', index=om.slicer[-1])

        prob.driver = NonSerDriver()

        prob.add_recorder(om.SqliteRecorder("cases1.sql"))
        prob.driver.add_recorder(om.SqliteRecorder("cases2.sql"))
        prob.model.add_recorder(om.SqliteRecorder("cases3.sql"))
        prob.model.nonlinear_solver.add_recorder(om.SqliteRecorder("cases4.sql"))

        prob.setup()
        prob.set_val('nsi.problem', em_prob)

        prob.run_model()

        cr = om.CaseReader("cases1.sql")

        dval = cr.problem_metadata['tree']['children'][1]['options']['bad']
        key, val = [(k, v) for k, v in dval[0].items()][0]
        self.assertTrue("(1, <class" in key)
        self.assertEqual(val, [2, 3])

        self.assertEqual(cr.problem_metadata['tree']['children'][1]['options']['bad2'],
                         {'((1,), (2,))': 'stuff'})
        self.assertEqual(cr.problem_metadata['tree']['children'][1]['options']['nonrec'],
                         'Not Recordable')
        self.assertEqual(cr.problem_metadata['tree']['children'][1]['options']['cx'],
                         '(3+7j)')

        dval = cr.problem_metadata['tree']['children'][1]['children'][4]['val']
        key, val = [(k, v) for k, v in dval[0].items()][0]
        self.assertTrue("Discrete_o', <class" in key)
        self.assertEqual(val, [2, {'((1,), (2,))': 'stuff'}])

        dval = cr.problem_metadata['tree']['children'][1]['children'][1]['val']
        key, val = [(k, v) for k, v in dval[0].items()][0]
        self.assertTrue("Discrete_i', <class" in key)
        self.assertEqual(val, [2, {'((1,), (2,))': 'stuff'}])

        self.assertEqual(cr.problem_metadata['tree']['children'][1]['children'][2]['val'],
                         '(3+5j)')


if __name__ == "__main__":
    unittest.main()



"""
jax implementations of the Kreisselmeier-Steinhauser for the min and max values in an array.
"""

try:
    import jax
    from jax import jit
    import jax.numpy as jnp
    jax.config.update("jax_enable_x64", True)
except (ImportError, ModuleNotFoundError):
    jax = None
    from ..utils.jax_utils import jit_stub as jit

CITATIONS = """
@conference {Martins:2005:SOU,
        title = {On Structural Optimization Using Constraint Aggregation},
        booktitle = {Proceedings of the 6th World Congress on Structural and Multidisciplinary
                     Optimization},
        year = {2005},
        month = {May},
        address = {Rio de Janeiro, Brazil},
        author = {Joaquim R. R. A. Martins and Nicholas M. K. Poon}
}
"""


@jit
def ks_max(x, rho=100.0):
    """
    Compute a differentiable maximum value in an array.

    Given some array of values `x`, compute a differentiable, _conservative_ maximum using the
    Kreisselmeier-Steinhauser function.

    Parameters
    ----------
    x : ndarray
        Array of values.
    rho : float
        Aggregation Factor. Larger values of rho more closely match the true maximum value.

    Returns
    -------
    float
        A conservative approximation to the minimum value in x.
    """
    x_max = jnp.max(x)
    x_diff = x - x_max
    exponents = jnp.exp(rho * x_diff)
    summation = jnp.sum(exponents)
    return x_max + 1.0 / rho * jnp.log(summation)


@jit
def ks_min(x, rho=100.0):
    """
    Compute a differentiable minimum value in an array.

    Given some array of values `x`, compute a differentiable,
    _conservative_ minimum using the Kreisselmeier-Steinhauser function.

    Parameters
    ----------
    x : ndarray
        Array of values.
    rho : float
        Aggregation Factor. Larger values of rho more closely match the true minimum value.

    Returns
    -------
    float
        A conservative approximation to the minimum value in x.
    """
    x_min = jnp.min(x)
    x_diff = x_min - x
    exponents = jnp.exp(rho * x_diff)
    summation = jnp.sum(exponents)
    return x_min - 1.0 / rho * jnp.log(summation)
